import numpy as np
from . import enmap, utils, powspec
try: from . import interpol
except ImportError: pass

####### Flat sky lensing #######

def lens_map(imap, grad_phi, order=3, mode="spline", border="cyclic", trans=False, deriv=False, h=1e-7):
	"""Lens map imap[{pre},ny,nx] according to grad_phi[2,ny,nx], where phi is the
	lensing potential, and grad_phi, which can be computed as enmap.grad(phi), simply
	is the coordinate displacement for each pixel. order, mode and border specify
	details of the interpolation used. See enlib.interpol.map_coordinates for details.
	If trans is true, the transpose operation is performed. This is NOT equivalent to
	delensing.

	If the same lensing field needs to be reused repeatedly, then higher efficiency
	can be gotten from calling displace_map directly with precomputed pixel positions."""
	# Converting from grad_phi to pix has roughly the same cost as calling displace_map.
	# So almost a factor 2 in speed can be won from calling displace_map directly.
	pos = imap.posmap() + grad_phi
	pix = imap.sky2pix(pos, safe=False)
	if not deriv:
		return displace_map(imap, pix, order=order, mode=mode, border=border, trans=trans)
	else:
		# displace_map deriv gives us ndim,{pre},ny,nx
		dlens_pix = displace_map(imap, pix, order=order, mode=mode, border=border, trans=trans, deriv=True)
		res = dlens_pix[0]*0
		pad = (slice(None),)+(None,)*(imap.ndim-2)+(slice(None),slice(None))
		for i in range(2):
			pos2 = pos.copy()
			pos2[i] += h
			pix2 = imap.sky2pix(pos2, safe=False)
			dpix = (pix2-pix)/h
			res += np.sum(dlens_pix * dpix[pad],0)
		return res

def delens_map(imap, grad_phi, nstep=3, order=3, mode="spline", border="cyclic"):
	"""The inverse of lens_map, such that delens_map(lens_map(imap, dpos), dpos) = imap
	for well-behaved fields. The inverse does not always exist, in which case the
	equation above will only be approximately fulfilled. The inverse is computed by
	iteration, with the number of steps in the iteration controllable through the
	nstep parameter. See enlib.interpol.map_coordinates for details on the other
	parameters."""
	grad_phi = delens_grad(grad_phi, nstep=nstep, order=order, mode=mode, border=border)
	return lens_map(imap, -grad_phi, order=order, mode=mode, border=border)

def delens_grad(grad_phi, nstep=3, order=3, mode="spline", border="cyclic"):
	"""Helper function for delens_map. Attempts to find the undisplaced gradient
	given one that has been displaced by itself."""
	alpha = grad_phi
	for i in range(nstep):
		alpha = lens_map(grad_phi, -alpha, order=order, mode=mode, border=border)
	return alpha

def displace_map(imap, pix, order=3, mode="spline", border="cyclic", trans=False, deriv=False):
	"""Displace map m[{pre},ny,nx] by pix[2,ny,nx], where pix indicates the location
	in the input map each output pixel should get its value from (float). The output
	is [{pre},ny,nx]."""
	if not deriv: omap = imap.copy()
	else:         omap = enmap.empty((2,)+imap.shape, imap.wcs, imap.dtype)
	if not trans:
		interpol.map_coordinates(imap, pix, omap, order=order, mode=mode, border=border, trans=trans, deriv=deriv)
	else:
		interpol.map_coordinates(omap, pix, imap, order=order, mode=mode, border=border, trans=trans, deriv=deriv)
	return omap

# Compatibility function. Not quite equivalent lens_map above due to taking phi rather than
# its gradient as an argument.
def lens_map_flat(cmb_map, phi_map):
	raw_pix  = cmb_map.pixmap() + enmap.grad_pix(phi_map)
	# And extract the interpolated values. Because of a bug in map_pixels with
	# mode="wrap", we must handle wrapping ourselves.
	npad = int(np.ceil(max(np.max(-raw_pix),np.max(raw_pix-np.array(cmb_map.shape[-2:])[:,None,None]))))
	pmap = enmap.pad(cmb_map, npad, wrap=True)
	return enmap.samewcs(utils.interpol(pmap, raw_pix+npad, order=4, mode="wrap"), cmb_map)

######## Curved sky lensing ########


def phi_to_kappa(phi_alm,phi_ainfo=None):
	"""Convert lensing potential alms phi_alm to
	lensing convergence alms kappa_alm, i.e.
	phi_alm * l * (l+1) / 2

	Args:
	    phi_alm: (...,N) ndarray of spherical harmonic alms of lensing potential
	    phi_ainfo: 	If ainfo is provided, it is an alm_info describing the layout 
	of the input alm. Otherwise it will be inferred from the alm itself.

	Returns:
	    kappa_alm: The filtered alms phi_alm * l * (l+1) / 2
	"""
	from . import curvedsky
	return curvedsky.almxfl(alm=phi_alm,lfilter=lambda l: l*(l+1)/2,ainfo=phi_ainfo)

def kappa_to_phi(kappa_alm,kappa_ainfo=None):
	"""Convert lensing convergence alms kappa_alm to
	lensing potential alms phi_alm, i.e.
	kappa_alm / ( l * (l+1) / 2 )

	Args:
	    kappa_alm: (...,N) ndarray of spherical harmonic alms of lensing convergence
	    kappa_ainfo: If ainfo is provided, it is an alm_info describing the layout 
	of the input alm. Otherwise it will be inferred from the alm itself.

	Returns:
	    phi_alm: The filtered alms kappa_alm / ( l * (l+1) / 2 )
	"""
	from . import curvedsky
	with utils.nowarn():
		oalm = curvedsky.almxfl(alm=kappa_alm,lfilter=lambda l: 1./(l*(l+1)/2) ,ainfo=kappa_ainfo)
	return utils.remove_nan(oalm)


def kappa_to_phi(kappa_alm,kappa_ainfo=None):
	"""Convert lensing convergence alms kappa_alm to
	lensing potential alms phi_alm, i.e.
	kappa_alm / ( l * (l+1) / 2 )

	Args:
	    kappa_alm: (...,N) ndarray of spherical harmonic alms of lensing convergence
	    kappa_ainfo: If ainfo is provided, it is an alm_info describing the layout 
	of the input alm. Otherwise it will be inferred from the alm itself.

	Returns:
	    phi_alm: The filtered alms kappa_alm / ( l * (l+1) / 2 )
	"""
	from . import curvedsky
	oalm = curvedsky.almxfl(alm=kappa_alm,lfilter=lambda x: 1./(x*(x+1)/2) ,ainfo=kappa_ainfo)
	oalm[~np.isfinite(oalm)] = 0
	return oalm


def lens_map_curved(shape, wcs, phi_alm, cmb_alm, phi_ainfo=None, maplmax=None, dtype=np.float64, spin=[0,2], output="l", geodesic=True, verbose=False, delta_theta=None):
	"""
	Lenses a CMB map given the lensing potential harmonic transform and the CMB 
	harmonic transform.  By default, T,E,B spherical harmonic coefficients are
	accepted and the returned maps are T, Q, U.
	Parameters:
	-----------
	shape : tuple
		Shape of the output map.
	wcs : WCS object
		World Coordinate System object describing the map projection.
	phi_alm : array-like
		Spherical harmonic coefficients of the lensing potential.
	cmb_alm : array-like
		Spherical harmonic coefficients of the CMB. If (3,nelem) shaped, the
		coeffients are assumed to be in the form of [T,E,B] in that order.
	phi_ainfo : alm_info, optional
		alm_info object containing information about the alm layout. Default: standard triangular layout.
	maplmax : int, optional
		Maximum multipole to use in the map.
	dtype : data-type, optional
		Data type of the output maps. Default is np.float64.
	spin : list, optional
		List of spins. These describe how to handle the [ncomp] axis in cmb_alm.
	 	0: scalar transform. Consumes one element in the component axis
	 	not 0: spin transform. Consumes two elements from the component axis.
	 	For example, if you have a TEB alm [3,nelem] and want to transform it
	 	to a TQU map [3,ny,nx], you would use spin=[0,2] to perform a scalar
	 	transform for the T component and a spin-2 transform for the Q,U
	 	components. Another example. If you had an alm [5,nelem] and map
	 	[5,ny,nx] and the first element was scalar, the next pair spin-1
	 	and the next pair spin-2, you woudl use spin=[0,1,2]. default:[0,2]
	output : str, optional
		String which specifies which maps to output, e.g. "lu". Default is "l".
		"l" - lensed CMB map
		"u" - unlensed CMB map
		"p" - lensing potential map
		"k" - convergence map
		"a" - deflection angle maps
	geodesic : bool, optional
		Properly parallel transport on the sphere (default). If False, a much faster
		approximation is used, which is still very accurate unless one is
		close to the poles.
	verbose : bool, optional
		If True, prints progress information. Default is False.
	delta_theta : float, optional
		Step size for the theta coordinate. 
	Returns:
	--------
	tuple
		A tuple containing the requested output maps in the order specified by the `output` parameter.
	"""

	from . import curvedsky
	# Restrict to target number of components
	oshape  = shape[-3:]
	if len(oshape) == 2: shape = (1,)+tuple(shape)
	if delta_theta is None: bsize = shape[-2]
	else:
		bsize = utils.nint(abs(delta_theta/utils.degree/wcs.wcs.cdelt[1]))
		# Adjust bsize so we don't get any tiny blocks at the end
		nblock= shape[-2]//bsize
		bsize = int(shape[-2]/(nblock+0.5))
	# Allocate output maps
	if "p" in output: phi_map   = enmap.empty(shape[-2:], wcs, dtype=dtype)
	if "k" in output:
		kappa_map = enmap.empty(shape[-2:], wcs, dtype=dtype)
		kappa_alm = phi_to_kappa(phi_alm,phi_ainfo=phi_ainfo)
		for i1 in range(0, shape[-2], bsize):
			curvedsky.alm2map(kappa_alm, kappa_map[...,i1:i1+bsize,:])
		del kappa_alm
	if "a" in output: grad_map  = enmap.empty((2,)+shape[-2:], wcs, dtype=dtype)
	if "u" in output: cmb_raw   = enmap.empty(shape, wcs, dtype=dtype)
	if "l" in output: cmb_obs   = enmap.empty(shape, wcs, dtype=dtype)
	# Then loop over dec bands
	for i1 in range(0, shape[-2], bsize):
		i2 = min(i1+bsize, shape[-2])
		lshape, lwcs = enmap.slice_geometry(shape, wcs, (slice(i1,i2),slice(None)))
		if "p" in output:
			if verbose: print("Computing phi map")
			curvedsky.alm2map(phi_alm, phi_map[...,i1:i2,:])
		if verbose: print("Computing grad map")
		if "a" in output: grad = grad_map[...,i1:i2,:]
		else: grad = enmap.zeros((2,)+lshape[-2:], lwcs, dtype=dtype)
		curvedsky.alm2map(phi_alm, grad, deriv=True)
		if "l" not in output: continue
		if verbose: print("Computing observed coordinates")
		obs_pos = enmap.posmap(lshape, lwcs)
		if verbose: print("Computing alpha map")
		raw_pos = enmap.samewcs(offset_by_grad(obs_pos, grad, pol=shape[-3]>1, geodesic=geodesic), obs_pos)
		del obs_pos, grad
		if "u" in output:
			if verbose: print("Computing unlensed map")
			curvedsky.alm2map(cmb_alm, cmb_raw[...,i1:i2,:], spin=spin)
		if verbose: print("Computing lensed map")
		cmb_obs[...,i1:i2,:] = curvedsky.alm2map_pos(cmb_alm, raw_pos[:2], spin=spin)
		if raw_pos.shape[0] > 2 and np.any(raw_pos[2]):
			if verbose: print("Rotating polarization")
			cmb_obs[...,i1:i2,:] = enmap.rotate_pol(cmb_obs[...,i1:i2,:], raw_pos[2])
		del raw_pos

	del cmb_alm, phi_alm
	# Output in same order as specified in output argument
	res = []
	for c in output:
		if   c == "l": res.append(cmb_obs.reshape(oshape))
		elif c == "u": res.append(cmb_raw.reshape(oshape))
		elif c == "p": res.append(phi_map)
		elif c == "k": res.append(kappa_map)
		elif c == "a": res.append(grad_map)
	return tuple(res)

def rand_alm(ps_lensinput, lmax=None, dtype=np.float64, seed=None, phi_seed=None, verbose=False, ncomp=None):
	from . import curvedsky
	ctype   = np.result_type(dtype,0j)
	if ncomp is not None: ps_lensinput = ps_lensinput[:1+ncomp,:1+ncomp]
	# First draw a random lensing field, and use it to compute the undeflected positions
	if verbose: print("Generating alms")
	if phi_seed is None:
		alm, ainfo = curvedsky.rand_alm(ps_lensinput, lmax=lmax, seed=seed, dtype=ctype, return_ainfo=True)
	else:
		# We want separate seeds for cmb and phi. This means we have to do things a bit more manually
		wps, ainfo = curvedsky.prepare_ps(ps_lensinput, lmax=lmax)
		alm = np.empty([1+ncomp,ainfo.nelem],ctype)
		curvedsky.rand_alm_white(ainfo, alm=alm[:1], seed=phi_seed)
		curvedsky.rand_alm_white(ainfo, alm=alm[1:], seed=seed)
		ps12 = enmap.multi_pow(wps, 0.5)
		ainfo.lmul(alm, (ps12/2**0.5).astype(dtype), alm)
		alm[:,:ainfo.lmax].imag  = 0
		alm[:,:ainfo.lmax].real *= 2**0.5
		del wps, ps12
		
	# Return phi_alm and cmb_alm
	return alm[0], alm[1:], ainfo


def rand_map(shape, wcs, ps_lensinput, lmax=None, maplmax=None, dtype=np.float64, seed=None, phi_seed=None, spin=[0,2], output="l", geodesic=True, verbose=False, delta_theta=None):
	# Restrict to target number of components
	oshape  = shape[-3:]
	if len(oshape) == 2: shape = (1,)+tuple(shape)
	ncomp   = shape[-3]
	phi_alm, cmb_alm, ainfo = rand_alm(ps_lensinput=ps_lensinput,
		lmax=lmax, dtype=dtype, seed=seed, phi_seed=phi_seed, verbose=verbose, ncomp=ncomp)

	# Truncate alm if we want a smoother map. In taylens, it was necessary to truncate
	# to a lower lmax for the map than for phi, to avoid aliasing. The appropriate lmax
	# for the cmb was the one that fits the resolution. FIXME: Can't slice alm this way.
	#if maplmax: cmb_alm = cmb_alm[:,:maplmax]
	return lens_map_curved(shape=shape, wcs=wcs, phi_alm=phi_alm,
		cmb_alm=cmb_alm, phi_ainfo=ainfo, maplmax=maplmax,
		dtype=dtype, spin=spin,
		output=output, geodesic=geodesic, verbose=verbose,
		delta_theta=delta_theta)

def offset_by_grad(ipos, grad, geodesic=True, pol=None):
	"""Given a set of coordinates ipos[{dec,ra},...] and a gradient
	grad[{ddec,dphi/cos(dec)},...] (as returned by curvedsky.alm2map(deriv=True)),
	returns opos = ipos + grad, while properly parallel transporting
	on the sphere. If geodesic=False is specified, then an much faster
	approximation is used, which is still very accurate unless one is
	close to the poles."""
	ncomp = 2 if pol is False or pol is None and ipos.shape[0] <= 2 else 3
	opos = np.empty((ncomp,)+ipos.shape[1:])
	iflat = ipos.reshape(ipos.shape[0],-1)
	# Oflat is a flattened view of opos, so changes to oflat
	# are visible in our return value opos
	oflat = opos.reshape(opos.shape[0],-1)
	gflat = grad.reshape(grad.shape[0],-1)
	if geodesic:
		# Loop over chunks in order to conserve memory
		step = 0x100000
		for i in range(0, iflat.shape[1], step):
			# The helper function assumes zenith coordinates
			small_grad = gflat[:,i:i+step].copy(); small_grad[0] = -small_grad[0]
			small_ipos = iflat[:,i:i+step].copy(); small_ipos[0] = np.pi/2-small_ipos[0]
			# Compute the offset position and polarization rotation, the latter in the
			# form of cos and sin of the polarization rotation angle
			small_opos, small_orot = offset_by_grad_helper(small_ipos, small_grad, ncomp>2)
			oflat[0,i:i+step] = np.pi/2 - small_opos[0]
			oflat[1,i:i+step] = small_opos[1]
			# Handle rotation if necessary
			if oflat.shape[0] > 2:
				# Recover angle from cos and sin
				oflat[2,i:i+step] = np.arctan2(small_orot[1],small_orot[0])
				if iflat.shape[0] > 2:
					oflat[2,i:i+step] += iflat[2,i:i+step]
	else:
		oflat[0] = iflat[0] + gflat[0]
		oflat[1] = iflat[1] + gflat[1]/np.cos(iflat[0])
		oflat[:2] = pole_wrap(oflat[:2])
		if oflat.shape[0] > 2: oflat[2] = 0
	return opos

def offset_by_grad_helper(ipos, grad, pol):
	"""Find the new position and induced rotation
	from offseting the input positions ipos[2,nsamp]
	by grad[2,nsamp]."""
	grad = np.array(grad)
	# Decompose grad into direction and length.
	# Fix zero-length gradients first, to avoid
	# division by zero.
	grad[:,np.all(grad==0,0)] = 1e-20
	d = np.sum(grad**2,0)**0.5
	grad  /=d
	# Perform the position offset using spherical
	# trigonometry
	cosd, sind = np.cos(d), np.sin(d)
	cost, sint = np.cos(ipos[0]), np.sin(ipos[0])
	ocost  = cosd*cost-sind*sint*grad[0]
	osint  = (1-ocost**2)**0.5
	ophi   = ipos[1] + np.arcsin(sind*grad[1]/osint)
	if not pol:
		return np.array([np.arccos(ocost), ophi]), None
	# Compute the induced polarization rotation.
	with utils.nowarn():
		# If sint is zero then we get a divide by zero here, but
		# the final result is still finite
		A    = grad[1]/(sind*cost/sint+grad[0]*cosd)
	nom1   = grad[0]+grad[1]*A
	denom  = 1+A**2
	cosgam = 2*nom1**2/denom-1
	singam = 2*nom1*(grad[1]-grad[0]*A)/denom
	res = np.array([np.arccos(ocost), ophi]), np.array([cosgam,singam])
	return res

def pole_wrap(pos):
	"""Handle pole wraparound."""
	a = np.array(pos)
	bad = np.where(a[0] > np.pi/2)
	a[0,bad] = np.pi - a[0,bad]
	a[1,bad] = a[1,bad]+np.pi
	bad = np.where(a[0] < -np.pi/2)
	a[0,bad] = -np.pi - a[0,bad]
	a[1,bad] = a[1,bad]+np.pi
	return a



"""This module defines shortcuts for generating WCS instances and working
with them. The bounding boxes and shapes used in this module all use
the same ordering as WCS, i.e. column major (so {ra,dec} rather than
{dec,ra}). Coordinates are assigned to pixel centers, as WCS does natively,
but bounding boxes include the whole pixels, not just their centers, which
is where the 0.5 stuff comes from."""
import numpy as np, warnings
from astropy.wcs import WCS, FITSFixedWarning

# Turn off annoying warning every time a WCS object is constructed
try:
	warnings.filterwarnings("ignore", category=FITSFixedWarning)
except AssertionError:
	# This try/catch is a hack for readthedocs builds.
	pass

# Handle annoying python3 stuff
try: basestring
except: basestring = str
def streq(x, s): return isinstance(x, basestring) and x == s

# Geometry construction redesign
#
# The old approach was build around the reference point. The idea was that
# this point would always be a pixel center, no matter which coutout of
# the sky one was looking at. The problem with this approach is that it doesn't
# generalize to downgrading, and it clashes with finer detalies such as
# distinguishing between CC and Fejer1, which care about the pixel alignment
# at the poles, not the equator where the reference point usually is.
#
# The new approach will proceed in three steps:
# 1. Specify the projection (ctype, crval) without any pixel details
# 2. Turn it into a full-sky pixelization (crpix, cdelt). This could
#    be done by specifying ny,nx or the resolution. We could here
#    issue a warning or exception if the sky isn't evenly tiled.
#    This part would care about sub-specifiers like :cc or :fejer1
# 3. Crop this to cover the target area
#
# These can all be handled in separate functions. The output from step
# 1 would be a wcs with default crpix and cdelt values.
#
# Problems:
# 1. Currently pixelization can't handle all of these:
#    * Fix left side but allow right side to float
#    * Fix right side but allow left side to float
#    * Fix both sides, but allow total width to float
#    Right now the first two are supported, but not the last.
#    For example, for CEA we can't expect a reasonable resolution
#    to reach the poles with a senible pixel offset, but we want to
#    at least make things symmetric around the equator. There's a choice
#    between trying to get a pixel edge as close to the poles as possible
#    or trying to get a pixel center as close to the poles as possible.
#    These could be written as hh adjust and 00 adjust, but hard to fit
#    this in currently.
# 2. Some projections have extra parameters, like CEA with lambda and
#    ZEA where one might want to be locally conformal. The approach where
#    one first builds the fullsky geometry and only later worried about
#    restricting it to a part of the sky clashes with this. Can support it,
#    but is it a good idea to do it automatically?

def projection(system, crval=None):
	"""Generate a pixelization-agnostic wcs"""
	system = system.lower()
	if crval is None: crval = default_crval(system)
	else: crval = np.zeros(2)+crval
	if system in ["", "plain"]: return explicit(crval=crval)
	return explicit(ctype=["RA---"+system.upper(), "DEC--"+system.upper()], crval=crval)

def pixelization(pwcs, shape=None, res=None, variant=None):
	"""Add pixel information to a wcs, returning a full-sky geometry,
	or as close to that as the projection allows."""
	# This is the hard part. Many projections have invalid areas, and
	# some have infinite size. May just have to handle the cases one by
	# one instead of trying to be general
	system   = get_proj(pwcs)
	extent, lonpole   = default_extent(system)
	variant  = variant or default_variant(system)
	offs     = parse_variant(variant)
	periodic = is_periodic(system)
	# We will now split our extent into pixels. Find the intermediate
	# coordinates of the first and last pixel center along each axis
	if shape is None:
		res = expand_res(res)
		ra1, ra2, nx, ox1, ox2 = pixelize_1d(extent[0], res=res[0],  offs=offs[0], periodic=periodic[0])
		dec1,dec2,ny, oy1, oy2 = pixelize_1d(extent[1], res=res[1],  offs=offs[1], periodic=periodic[1])
	elif res is None:
		ra1, ra2, nx, ox1, ox2 = pixelize_1d(extent[0], n=shape[-2], offs=offs[0], periodic=periodic[0])
		dec1,dec2,ny, oy1, oy2 = pixelize_1d(extent[1], n=shape[-2], offs=offs[1], periodic=periodic[0])
	else:
		raise ValueError("Either res or shape must be given to build a pixelization")
	# Now that we have the intermediate coordinates of our endpoints, we
	# can calculate cdelt and crpix
	owcs  = pwcs.deepcopy()
	owcs.wcs.cdelt = [(ra2-ra1)/(nx-1), (dec2-dec1)/(ny-1)]
	# The bottom-left corner has pixel coordinates -ox1,-oy1
	# The top-right   corner has pixel coordinates (nx-1)-ox2,(ny-1)-oy2
	# The center is the average of these
	owcs.wcs.crpix[0] = 1+((nx-1)-ox2-ox1)/2
	owcs.wcs.crpix[1] = 1+((ny-1)-oy2-oy1)/2
	if lonpole is not None:
		owcs.wcs.lonpole = lonpole
	return (ny,nx), owcs

def explicit(naxis=2, **args):
	wcs = WCS(naxis=naxis)
	for key in args:
		setattr(wcs.wcs, key, args[key])
	return wcs

def expand_res(res, signs=None, flip=False):
	"""If res is not None, expand it to length 2. If it wasn't already
	length 2, the RA sign will be inverted. If flip is True, the res order
	will be flipped before expanding"""
	if res is None: return res
	# Bleh, compensate for later flip
	if signs is None: signs = [1,-1] if flip else [-1,1]
	res = np.atleast_1d(res)
	assert res.ndim == 1, "Invalid res shape"
	assert len(res) <= 2, "Invalid res length"
	if flip: res, signs = res[::-1], signs[::-1]
	if res.size == 1: res = np.array(signs)*res[0]
	return res

def describe(wcs):
	"""Since astropy.wcs.WCS objects do not have a useful
	str implementation, this function provides a relpacement."""
	sys  = wcs.wcs.ctype[0][-3:].lower()
	n    = wcs.naxis
	fields = ("cdelt:["+",".join(["%.4g"]*n)+"],crval:["+",".join(["%.4g"]*n)+"],crpix:["+",".join(["%.2f"]*n)+"]") % (tuple(wcs.wcs.cdelt) + tuple(wcs.wcs.crval) + tuple(wcs.wcs.crpix))
	pv = wcs.wcs.get_pv()
	for p in pv:
		fields += ",pv[%d,%d]=%.3g" % p
	return "%s:{%s}" % (sys, fields)
# Add this to all WCSes in this class
WCS.__repr__ = describe
WCS.__str__ = describe

def equal(wcs1, wcs2,flags=1,tol=1e-14):
	return wcs1.wcs.compare(wcs2.wcs, flags, tol)

def nobcheck(wcs):
	res = wcs.deepcopy()
	res.wcs.bounds_check(False, False)
	return res

def is_compatible(wcs1, wcs2, tol=1e-3):
	"""Checks whether two world coordinate systems represent
	(shifted) versions of the same pixelizations, such that
	every pixel center in wcs1 correspond to a pixel center in
	wcs2. For now, they also have to have the pixels going
	in the same direction."""
	h1 = wcs1.to_header()
	h2 = wcs2.to_header()
	keys = sorted(list(set(h1.keys())&set(h2.keys())))
	for key in keys:
		if key.startswith("CRVAL") or key.startswith("CRPIX") or key.startswith("CDELT"): continue
		if key not in h2 or h2[key] != h1[key]: return False
	if np.max(np.abs(wcs1.wcs.cdelt-wcs2.wcs.cdelt))/np.min(np.abs(wcs1.wcs.cdelt)) > tol: return False
	crdelt = wcs1.wcs.crval - wcs2.wcs.crval
	cpdelt = wcs1.wcs.crpix - wcs2.wcs.crpix
	subpix = (crdelt/wcs1.wcs.cdelt - cpdelt + 0.5)%1-0.5
	if np.max(np.abs(subpix)) > tol: return False
	return True

def is_plain(wcs):
	"""Determines whether the given wcs represents plain, non-specific,
	non-wrapping coordinates or some angular coordiante system."""
	return get_proj(wcs) in ["","plain"]

def is_cyl(wcs):
	"""Returns True if the wcs represents a cylindrical coordinate system"""
	return get_proj(wcs) in ["cyp","cea","car","mer"]

def is_separable(wcs):
	return is_cyl(wcs) and wcs.wcs.crval[1] == 0

def get_proj(wcs):
	if isinstance(wcs, str): return wcs
	else:
		toks = wcs.wcs.ctype[0].split("-")
		return toks[-1].lower() if len(toks) >= 2 else ""

def parse_system(system, variant=None):
	toks = system.split(":")
	if len(toks) > 1: return toks[0].lower(), toks[1]
	else: return toks[0].lower(), variant

def scale(wcs, scale=1, rowmajor=False, corner=True):
	"""Scales the linear pixel density of a wcs by the given factor, which can be specified
	per axis. This is the same as dividing the pixel size by the same numberr
	corner controls which area is scaled. With corner=True (the default), then the
	area from the start of the first pixel to the end of the lats pixel will be scaled
	by this factor. If corner=False, then the area from the center of the first pixel
	to the center of the last pixel will be scaled. Usually the former makes most sense."""
	scale = np.zeros(2)+scale
	if rowmajor: scale = scale[::-1]
	wcs = wcs.deepcopy()
	if corner:
		wcs.wcs.crpix -= 0.5
	wcs.wcs.crpix *= scale
	wcs.wcs.cdelt /= scale
	if corner:
		wcs.wcs.crpix += 0.5
	return wcs

#def expand_res(res, default_dirs=[1,-1]):
#	res = np.atleast_1d(res)
#	assert res.ndim == 1, "Invalid res shape"
#	if res.size == 1:
#		return np.array(default_dirs)*res
#	else:
#		return res

###########################
#### Helper functions #####
###########################

def is_azimuthal(system): return system.lower() in ["arc", "zea", "sin", "tan", "azp", "slp", "stg", "zpn", "air"]

def default_crval(system):
	if is_azimuthal(system): return [0,90]
	else: return [0,0]

def default_extent(system):
	"""Return the horizontal and vertical extent of the full sky in
	degrees, and the prefered value of lonpole (or None if it should
	be left alone).  For some systems the full sky is not
	representable, in which case a reasonable compromise is returned

	"""
	system = system.lower()
	if   system in ["", "plain"]: return [1,1], None
	# Cylindrical
	if   system == "car": return [360,180], None
	elif system == "cea": return [360,360/np.pi], None
	elif system == "mer": return [360,360], None # traditional dec range gives square map
	# Zenithal
	elif system == "arc": return [360,360], 180.
	elif system == "zea": return [720/np.pi,720/np.pi], 180.
	elif system == "sin": return [360/np.pi,360/np.pi], 180. # only orthographic supported
	elif system == "tan": return [360,360], 180. # goes down to 0.158° above the horizon
	# Pseudo-cyl
	elif system == "mol": return [720*2**0.5/np.pi,360*2**0.5/np.pi], None
	elif system == "ait": return [720*2**0.5/np.pi,360*2**0.5/np.pi], None
	else: raise ValueError("Unsupported system '%s'" % str(system))

def default_variant(system):
	system = system.lower()
	return "fejer1" if system in ["car","plain",""] else "any"

def extent2bounds(extent): return [[-e/h,e/h] for e in extent]

def is_periodic(system):
	system = system.lower()
	if is_azimuthal(system) or system in ["", "plain"]:
		return [False,False]
	else:
		return [True,False]

def parse_variant(name):
	name = name.lower()
	if   name == "safe":   rule = "hh,hh" # fully-downgrade safe. What fejer1 should have been
	elif name == "fejer1": rule = "00,hh" # stays SHTable after downgrade, but not pix-comp with raw @ that res
	elif name == "cc":     rule = "00,00" # what we used for pre-DR6. Cannot SHT after downgrade
	elif name == "any":    rule = "**,**"
	else: rule = name
	toks = rule.split(",")
	if len(toks) != 2 or len(toks[0]) != 2 or len(toks[1]) != 2:
		raise ValueError("Could not recognize pixelization variant '%s'" % (str(name)))
	left  = {"0": 0, "h": 0.5, "*": None}
	right = {"0": 0, "h":-0.5, "*": None}
	try:
		return [[left[tok[0]],right[tok[1]]] for tok in toks]
	except KeyError:
		raise ValueError("Invalid character in rule '%s'" % str(rule))

class PixelizationError(Exception): pass

def pixelize_1d(w, n=None, res=None, offs=None, periodic=False, adjust=False, sign=1, tol=1e-6, eps=1e-6):
	"""Figure out how to align pixels along an interval w long such
	that there are either n pixels or the resolution is res, and with
	the given pixel offsets from the edges. Returns the coordinates of
	the center of the first and last pixel."""
	# FIXME: This is a bit poorly thought out. The concept of being
	# able to adjust the range and that of having wildcard edges should
	# be separate, but the way we've done things now there's no room in
	# parse_variant to say something like "0h" but adjustable. For now
	# I just have to hardcode that "**" means "00" but adjustable.
	o1, o2 = offs if offs is not None else (None, None)
	if res is not None:
		if res < 0: res, sign = -res, -sign
		if o1 is None and o2 is None:
			o1 = o2 = 0
			adjust = True
		if o2 is None:
			# Add a tiny number to avoid having a rounding discontinuity for common values
			# off w, res and o1
			n   = int(w/res+1-o1+eps)
		elif o1 is None:
			n   = int(w/res+1+o2+eps)
		else:
			# Both given! Can we satisfy requirement?
			nf = w/res+1-(o1-o2)
			n  = int(nf+eps)
			if adjust:
				# We're free to redefine w so things work
				w = (n+(o1+o2)-1)*res
			else:
				# Complain if the resolution and offsets are incompatible
				if not np.abs(n-nf)<tol:
					raise PixelizationError("Resolution %g does not evenly divide extent %g with offsets [%g,%g]" %
						(res, w, o1, o2))
	else:
		if o1 is None: o1 =  0.5
		if o2 is None: o2 = -0.5
		res = w/(n-1+o1-o2)
	# Finish up
	if o1 is not None:
		ra1 = -w/2+o1*res
		ra2 = ra1+(n-1)*res
	else:
		ra2 = +w/2+o2*res
		ra1 = ra2 - (n-1)*res
	# If this axis is periodic, then the last point could be equal to the first,
	# depending on the pixelization
	if periodic and np.allclose(ra2-ra1,w):
		ra2 -= res
		n   -= 1
	# Apply any sign
	ra1 *= sign
	ra2 *= sign
	return ra1, ra2, n, o1, o2

def recenter_cyl_x(wcs, x):
	"""Given a cylindrical wcs with the reference point already on the equator,
	move the reference point along the equator to the given x (counting from 1)
	returning a new wcs."""
	if not is_separable(wcs):
		raise ValueError("recenter_cyl requires a cylindrical wcs with crval on the equator")
	owcs = wcs.deepcopy()
	owcs.wcs.crpix[0]  = x
	owcs.wcs.crval[0] += (x-wcs.wcs.crpix[0])*wcs.wcs.cdelt[0]
	return owcs

def recenter_cyl_ra(wcs, ra):
	return recenter_cyl_x(wcs.wcs.crpix[0] + (ra-wcs.wcs.crval[0])/wcs.wcs.cdelt[0])

def fix_wcs(wcs, axis=0):
	"""Returns a new WCS object which has had the reference pixel moved to the
	middle of the possible pixel space."""
	res = wcs.deepcopy()
	# Find the center ra manually: mean([crval - crpix*cdelt, crval + (-crpix+shape)*cdelt])
	#  = crval + (-crpix+shape/2)*cdelt
	# What pixel does this correspond to?
	#  crpix2 = crpix + (crval2-crval)/cdelt
	# But that requires shape. Can we do without it? Yes, let's use the
	# biggest possible shape. n = 360/cdelt
	n = abs(360/wcs.wcs.cdelt[axis])
	delta_ra  = wcs.wcs.cdelt[axis]*(n/2-wcs.wcs.crpix[axis])
	delta_pix = delta_ra/wcs.wcs.cdelt[axis]
	res.wcs.crval[axis] += delta_ra
	res.wcs.crpix[axis] += delta_pix
	repr(res.wcs) # wcs not properly updated if I don't do this
	return res

def fix_cdelt(wcs):
	"""Return a new wcs with pc and cd replaced by cdelt"""
	owcs = wcs.deepcopy()
	if wcs.wcs.has_cd():
		del owcs.wcs.cd, owcs.wcs.pc
		owcs.wcs.cdelt *= np.diag(wcs.wcs.cd)
	elif wcs.wcs.has_pc():
		del owcs.wcs.cd, owcs.wcs.pc
		owcs.wcs.cdelt *= np.diag(wcs.wcs.pc)
	return owcs

# The functions below are used to implement the old
# patch-oriented geometry functions.

# The origin argument used in the wcs pix<->world routines seems to
# have to be 1 rather than the 0 one would expect. For example,
# if wcs is CAR(crval=(0,0),crpix=(0,0),cdelt=(1,1)), then
# pix2world(0,0,1) is (0,0) while pix2world(0,0,0) is (-1,-1).
#
# No! the problem is that everythin in the fits header counts from 1,
# so the default crpix should be (1,1), not (0,0). With
# CAR(crval(0,0),crpix(1,1),cdelt(1,1)) we get
# pix2world(1,1,1) = (0,0) and pix2world(0,0,0) = (0,0)

# Useful stuff to be able to do:
#  * Create a wcs from (point,res)
#  * Create a wcs from (box,res)
#  * Create a wcs from (box,shape)
#  * Create a wcs from (point,res,shape)
# Can support this by taking arguments:
#  pos: point[2] or box[2,2], mandatory
#  res: num or [2], optional
#  shape: [2], optional
# In cases where shape is not specified, the implied
# shape can be recovered from the wcs and a box by computing
# the pixel coordinates of the corners. So we don't need to return
# it.

#  1. Construct wcs from box, res (and return shape?)
#  2. Construct wcs from box, shape
#  3. Construct wcs from point, res (this is the most primitive version)

# I need to update this to work better with full-sky stuff.
# Should be easy to construct something that's part of a
# clenshaw-curtis or fejer sky.

deg2rad = np.pi/180
rad2deg = 1/deg2rad

def plain(pos, res=None, shape=None, rowmajor=False, ref=None):
	"""Set up a plain coordinate system (non-cyclical)"""
	pos, res, shape, mid = validate(pos, res, shape, rowmajor, default_dirs=[1,1])
	w = WCS(naxis=2)
	w.wcs.crval = mid
	if streq(ref, "standard"): ref = None
	return finalize(w, pos, res, shape, ref=ref)

def car(pos, res=None, shape=None, rowmajor=False, ref=None):
	"""Set up a plate carree system. See the build function for details."""
	pos, res, shape, mid = validate(pos, res, shape, rowmajor)
	w = WCS(naxis=2)
	w.wcs.ctype = ["RA---CAR", "DEC--CAR"]
	w.wcs.crval = np.array([mid[0],0])
	if streq(ref, "standard"): ref = (0,0)
	return finalize(w, pos, res, shape, ref=ref)

def cea(pos, res=None, shape=None, rowmajor=False, lam=None, ref=None):
	"""Set up a cylindrical equal area system. See the build function for details."""
	pos, res, shape, mid = validate(pos, res, shape, rowmajor)
	if lam is None:
		lam = np.cos(mid[1]*deg2rad)**2
	w = WCS(naxis=2)
	w.wcs.ctype = ["RA---CEA", "DEC--CEA"]
	w.wcs.set_pv([(2,1,lam)])
	w.wcs.crval = np.array([mid[0],0])
	if streq(ref, "standard"): ref = (0,0)
	return finalize(w, pos, res, shape, ref=ref)

def mer(pos, res=None, shape=None, rowmajor=False, ref=None):
	"""Set up a mercator system. See the build function for details."""
	pos, res, shape, mid = validate(pos, res, shape, rowmajor)
	w = WCS(naxis=2)
	w.wcs.ctype = ["RA---MER", "DEC--MER"]
	w.wcs.crval = np.array([mid[0],0])
	if streq(ref, "standard"): ref = (0,0)
	return finalize(w, pos, res, shape, ref=ref)

def arc(pos, res=None, shape=None, rowmajor=False, ref=None):
	"""Setups up a zenithal equidistant projection.  See the build
	function for details.

	"""
	pos, res, shape, mid = validate(pos, res, shape, rowmajor)
	w = WCS(naxis=2)
	w.wcs.ctype = ["RA---ARC", "DEC--ARC"]
	w.wcs.crval = mid
	w, ref = _apply_zenithal_ref(w, ref)
	return finalize(w, pos, res, shape, ref=ref)

def sin(pos, res=None, shape=None, rowmajor=False, ref=None):
	"""Setups up an orthographic projection.  See the build function
	for details.

	"""
	pos, res, shape, mid = validate(pos, res, shape, rowmajor)
	w = WCS(naxis=2)
	w.wcs.ctype = ["RA---SIN", "DEC--SIN"]
	w.wcs.crval = mid
	w, ref = _apply_zenithal_ref(w, ref)
	return finalize(w, pos, res, shape, ref=ref)

def zea(pos, res=None, shape=None, rowmajor=False, ref=None):
	"""Setups up an oblate Lambert's azimuthal equal area system.
	See the build function for details. Don't use this if you want
	a polar projection."""
	pos, res, shape, mid = validate(pos, res, shape, rowmajor)
	w = WCS(naxis=2)
	w.wcs.ctype = ["RA---ZEA", "DEC--ZEA"]
	w.wcs.crval = mid
	w, ref = _apply_zenithal_ref(w, ref)
	return finalize(w, pos, res, shape, ref=ref)

# The airy distribution is a bit different, since is needs to
# know the size of the patch.
def air(pos, res=None, shape=None, rowmajor=False, rad=None, ref=None):
	"""Setups up an Airy system. See the build function for details."""
	pos, res, shape, mid = validate(pos, res, shape, rowmajor)
	if rad is None:
		if pos.ndim != 2:
			raise ValueError("Airy requires either rad or pos[2,2]")
		w = angdist(mid[0]*deg2rad,pos[0,1]*deg2rad,mid[0]*deg2rad,pos[1,1]*deg2rad)*rad2deg
		h = angdist(pos[0,0]*deg2rad,mid[1]*deg2rad,pos[1,0]*deg2rad,mid[1]*deg2rad)*rad2deg
		rad = (w+h)/4
	w = WCS(naxis=2)
	w.wcs.ctype = ["RA---AIR","DEC--AIR"]
	w.wcs.set_pv([(2,1,90-rad)])
	w, ref = _apply_zenithal_ref(w, ref)
	return finalize(w, pos, res, shape, ref=ref)

def tan(pos, res=None, shape=None, rowmajor=False, ref=None):
	"""Set up a gnomonic (tangent plane) system. See the build function for details."""
	pos, res, shape, mid = validate(pos, res, shape, rowmajor)
	w = WCS(naxis=2)
	w.wcs.ctype = ["RA---TAN", "DEC--TAN"]
	w.wcs.crval = mid
	w, ref = _apply_zenithal_ref(w, ref)
	return finalize(w, pos, res, shape, ref=ref)

systems = {"car": car, "cea": cea, "mer": mer, "air": air, "arc": arc, "sin": sin, "zea": zea, "tan": tan, "gnom": tan, "plain": plain }

def build(pos, res=None, shape=None, rowmajor=False, system="cea", ref=None, **kwargs):
	"""Set up the WCS system named by the "system" argument.
	pos can be either a [2] center position or a [{from,to},2]
	bounding box. At least one of res or shape must be specified.
	If res is specified, it must either be a number, in
	which the same resolution is used in each direction,
	or [2]. If shape is specified, it must be [2]. All angles
	are given in degrees."""
	return systems[system.lower()](pos, res, shape, rowmajor, ref=ref, **kwargs)

def validate(pos, res, shape, rowmajor=False, default_dirs=[1,-1]):
	pos = np.asarray(pos)
	if pos.shape != (2,) and pos.shape != (2,2):
		raise ValueError("pos must be [2] or [2,2]")
	if res is None and shape is None:
		raise ValueError("Atleast one of res and shape must be specified")
	if res is not None:
		res = np.atleast_1d(res)
		if res.shape == (1,):
			# If our shape has one entry, expand it to [y,x].
			# Two cases: 1. [2,2] pos given, in which case it has responsibility for
			# the coordinate directions, so we don't introduce a sign here, and
			# 2. [2] pos is given, in which case it's res's responsibility.
			if pos.shape == (2,2): res = np.zeros(2)+res
			else:                  res = np.array(default_dirs)*res
		elif res.shape != (2,):
			raise ValueError("res must be num or [2]")
	if rowmajor:
		pos = pos[...,::-1]
		if shape is not None: shape = shape[::-1]
		if res is not None: res = res[::-1]
	if shape is not None:
		shape = shape[:2]
	if res is None and pos.ndim != 2:
		raise ValueError("pos must be a bounding box if res is not specified")
	mid = pos if pos.ndim == 1 else np.mean(pos,0)
	return pos, res, shape, mid

def finalize(w, pos, res, shape, ref=None):
	"""Common logic for the various wcs builders. Fills in the reference
	pixel and resolution."""
	w.wcs.crpix = [1,1]
	if res is None:
		# Find the resolution that gives our box the required extent.
		w.wcs.cdelt = [1,1]
		corners = w.wcs_world2pix(pos,1)
		w.wcs.cdelt *= (corners[1]-corners[0])/shape
	else:
		w.wcs.cdelt = res
		if pos.ndim == 2: w.wcs.cdelt[pos[1]<pos[0]] *= -1
	if pos.ndim == 1:
		if shape is not None:
			# Place pixel origin at corner of shape centered on crval
			off = w.wcs_world2pix(pos[None],0)[0]
			w.wcs.crpix = np.array(shape)/2.0+0.5 - off
	else:
		# Make pos[0] the corner of the (0,0) pixel (counting from 0 for simplicity)
		off = w.wcs_world2pix(pos[0,None],0)[0]+0.5
		w.wcs.crpix -= off
	if ref is not None:
		# Tweak wcs so that crval is an integer number of
		# pixels away from ref.  This is most straight-forward
		# if one simply adjusts crpix.
		off = (w.wcs_world2pix(np.asarray(ref)[None], 1)[0] + 0.5) % 1 - 0.5
		w.wcs.crpix -= off
	return w

def _apply_zenithal_ref(w, ref):
	"""Input is a wcs w and ref is a position (dec,ra) or a special value
	(None, 'standard').  Returns tuple (w, ref_out).  If ref is a
	position, it is copied into w.wcs.crval and ref_out=ref.
	Otherwise, w is unmodified and ref_out=w.wcs.crval.  Also sets lonpole,
	if not already set, to 180, which is sensible default."""
	if np.isnan(w.wcs.lonpole):
		w.wcs.lonpole = 180.
	if isinstance(ref, str) and ref == 'standard':
		ref = None
	if ref is None:
		ref = w.wcs.crval
	else:
		w.wcs.crval = ref
	return w, ref

def angdist(lon1,lat1,lon2,lat2):
	return np.arccos(np.cos(lat1)*np.cos(lat2)*(np.cos(lon1)*np.cos(lon2)+np.sin(lon1)*np.sin(lon2))+np.sin(lat1)*np.sin(lat2))

def center_cyl_wcs(wcs, shape=None, off=0.5):
	"""Given the wcs for a cylindrical projection, return a new wcs
	where the reference point has been moved along the equator to
	the middle of the patch. This ensures that all pixels are within
	the standard WCS bounds. If shape is not passed, then the patch
	is assumed to cover the whole width of the sky."""
	# Can't manipulate crval if coordinates aren't separable
	if not is_separable(wcs):
		raise ValueError("Can't fix wcs for non-separable wcs")
	# Patch horizontal extent
	if shape is None: n = abs(360/wcs.wcs.cdelt[axis])
	else:             n = shape[-1]
	# x pixel of center  (1-based)
	x  = (n-1)/2+1
	# corresponding RA
	ra = wcs.wcs.crval[0] + (x-wcs.wcs.crpix[0])*wcs.wcs.cdelt[0]
	# We will allow negative RA, but we prefer small, positive values.
	# If a patch goes at most 180° away from crval, then we can require
	# 0 <= crval <= 360. But we add a small offset to not be sensitive to
	# tiny errors for our common choice of crval[0] = 0
	ra = (ra-off) % 360 + off
	# Make a new wcs with this
	owcs = wcs.deepcopy()
	owcs.wcs.crval[0] = ra
	owcs.wcs.crpix[0] = x
	return owcs

def fix_wcs(wcs, axis=0, n=None):
	"""Compatibility name for center_cyl_wcs"""
	if axis != 0: raise NotImplementedError
	return center_cyl_wcs(wcs, None if n is None else (1,n))

def fix_cdelt(wcs):
	"""Return a new wcs with pc and cd replaced by cdelt"""
	owcs = wcs.deepcopy()
	if wcs.wcs.has_cd():
		del owcs.wcs.cd, owcs.wcs.pc
		owcs.wcs.cdelt *= np.diag(wcs.wcs.cd)
	elif wcs.wcs.has_pc():
		del owcs.wcs.cd, owcs.wcs.pc
		owcs.wcs.cdelt *= np.diag(wcs.wcs.pc)
	return owcs



import unittest
import tempfile
import os

import numpy as np
import h5py
from pixell import enmap, utils

class IOTests(unittest.TestCase):
    def setUp(self):
        self.tempdir = tempfile.TemporaryDirectory()

    def tearDown(self):
        self.tempdir.cleanup()

    def _tempfile(self, basename):
        return os.path.join(self.tempdir.name, basename)

    def test_100_hdf(self):
        box = np.array([[-5,10],[5,-10]]) * utils.degree
        shape, wcs = enmap.geometry(
            pos=box, res=0.5 * utils.degree, proj='car')
        c_shape = (100, 200)

        super_filename = self._tempfile('test-super.h5')
        with h5py.File(super_filename, 'w') as super_out:
            filenames = []
            for name, super_shape in [
                    ('flat', ()),
                    ('classic', (3,)),
                    ('monster', (10, 23)),
            ]:           
                my_map = enmap.zeros(super_shape + c_shape,
                                     wcs=wcs)
                filenames.append(self._tempfile(f'test-{name}.h5'))
                enmap.write_hdf(filenames[-1], my_map)
                enmap.write_hdf(super_out, my_map, address=f'/{name}')

        # Single file read back...
        for f in filenames:
            map_in = enmap.read_hdf(f)
            print(f'File {f} contains a map with shape={map_in.shape}, wcs={map_in.wcs}')

        # Super file, with filename
        map_in = enmap.read_hdf(super_filename, address='monster')

        # Super file, with contextmanager
        with h5py.File(super_filename, 'r') as fin:
            for k in fin.keys():
                print(f'Found group {k} ...')
                map_in = enmap.read_hdf(fin, address=k)
                print(f'   {map_in.shape}, {map_in.wcs}')

if __name__ == '__main__':
    unittest.main()



"""Point source parameter I/O. In order to simulate a point source as it appears on
the sky, we need to know its position, amplitude and local beam shape (which can
also absorb an extendes size for the source, as long as it's gaussian). While other
properties may be nice to know, those are the only ones that matter for simulating
it. This module provides functions for reading these minimal parameters from
various data files.

The standard parameters are [nsrc,nparam]:
	dec (radians)
	ra (radians)
	[T,Q,U] amplitude at center of gaussian (uK)
	beam sigma (wide  axis) (radians)
	beam sigma (short axis) (radians)
	beam orientation (wide axis from dec axis)  (radians)

What do I really need to simulate a source?

1. Physical source on the sky (pos,amps,shape)
2. Telescope response (beam in focalplane)

For a point source 1.shape would be a point. But clusters and
nearby galaxies can have other shapes. In general many profiles are
possible. Parametrizing them in a standard format may be difficult.
"""
from __future__ import print_function, division
import numpy as np, time
from astropy.io import fits
from scipy import spatial
from . import utils, enmap, srcsim, wcsutils

#### Map-space source simulation ###

# New version. Usually 10 or more times faster than the old one, and with less memory
# overhead. But interface is a bit different, so it has a new name.
def sim_objects(shape, wcs, poss, amps, profile, prof_ids=None, omap=None, vmin=None, rmax=None,
		op="add", pixwin=False, pixwin_order=0, separable="auto", transpose=False, prof_equi="auto", cache=None,
		return_times=False):
	"""Simulate radially symmetric objects with arbitrary profiles and amplitudes.
	Arguments:
	* shape, wcs: The geometry of the patch to simulate. Only shape[-2:]
	  is used. amps determines the pre-dimensions
	* poss: The positions of the objects. [{dec,ra},nobj] in radians.
	* amps: The central amplitudes of the objects. [...,nobj]. Not the same as the flux.
	* profile: The profiles to use. Either [{r,b(r)},nsamp] (with shape (2,nsamp)) or a
	  list of such, where nsamp is the size of r and b(r).  If providing a list for
	  nobj objects, the shape of the array passed is (nobj,2,nsamp) and prof_ids
	  should be np.arange(nobj).

	Optional arguments:
	* prof_ids: Which profile to use for each source. Defaults to use
	  the first profile for all. Only necessary to specify if you want to
	  simulate objects with varying profiles. If specified, it should be
	  [nobj] indices into the profile list.
	* omap: Update this map instead of constructing a new one.
	  MUST BE float32 AND C CONTIGUOUS and have shape [...,ny,nx] where ... matches amps.
	* vmin: The lowest values to bother simulating. To avoid being terribly slow,
	  profiles aren't evaluated out to infinite distance, but only once they drop down
	  to a sufficiently low level, given by vmin. This takes into account the peak
	  amplitud of each object, so it should be in map units. For example, it might
	  be reasonable to have a vmin a few times lower than the noise level of the map,
	  e.g. 0.1 µK. If not specified, then it defaults to min(abs(amps))*1e-3.
	* rmax: The maximum radius to use, in radians. Acts as a cap on the radius
	  calculated from vmin. Not applied if None or 0.
	* op: What operation to use when combining the input map with each object.
	  "add": Add linearly [default]
	  "max": Keep the max value in each pixel
	  "min": Keep the min value in each pixel
	* pixwin: Whether to apply a pixel window after simulating the objects. This
	  assumes periodic boundary consitions, so objects at the very edge will be
	  wrong.
	* separable: Whether the coordinate system's coordinate axes are indpendent,
	  such that one only needs to know y in order to calculate dec, and x to
	  calculate ra. This allows for much faster calculation of the pixel
	  coordinates. Default "auto": True for cylindrical coordinates, False otherwise.
	* cache: Dictionary to use for caching pixel coordinates. Can be useful
	  if you're doing repeated simulations on the same geometry with non-separable
	  geometry, to avoid having to recalculate the pixel coordinates all the time.

	Returns the resulting map. If omap was specified, then the same object will
	be returned (after being updated of course). In this case, the simulated
	sources will have been added (or maxed etc. depending on op) into the map.
	Otherwise, the only signal in the map will be the objects."""
	dtype = np.float32 # C extension only supports this dtype
	if separable == "auto": separable = wcsutils.is_separable(wcs)
	# Object positions
	obj_decs = np.asanyarray(poss[0], dtype=dtype, order="C")
	obj_ras  = np.asanyarray(poss[1], dtype=dtype, order="C")
	obj_ys, obj_xs = utils.nint(enmap.sky2pix(shape, wcs, poss)).astype(np.int32)
	assert obj_decs.ndim == 1 and obj_ras.ndim == 1, "poss must be [{dec,ra},nobj]"
	nobj     = len(obj_decs)
	# Object amplitudes, and number of components
	amps     = np.asanyarray(amps, dtype=dtype, order="C")
	pre      = amps.shape[:-1]
	amps_flat= amps.reshape(-1, amps.shape[-1])
	ncomp    = len(amps_flat)
	# Profiles. For point sources this would just be the beam, but extended objects
	# can have differnet profiles. We will support both [{r,b}] and [[{r,b}],[{r,b}],...]
	try: profile[0][0][0]
	except (TypeError, IndexError): profile = [profile]
	profile = [np.asanyarray(p, dtype=dtype, order="C") for p in profile]
	# If which profile to use isn't specified, default to the first one
	if prof_ids is None: prof_ids = np.zeros(nobj, np.int32)
	else: prof_ids = np.asanyarray(prof_ids, dtype=np.int32, order="C")
	if prof_equi == "auto": prof_equi = all([is_equi(prof[0]) for prof in profile])
	# If user hasn't specified how faint things to simulate, then set the limit a bit
	# below the peak of the faintest object
	if vmin is None: vmin = np.min(np.abs(amps))*1e-3
	if rmax is None: rmax = 0
	# Set up the pixel coordinates
	if separable: posmap = utils.cache_get(cache, "posmap", lambda: enmap.posaxes(shape, wcs, dtype=dtype))
	else:         posmap = utils.cache_get(cache, "posmap", lambda: enmap.posmap (shape, wcs, dtype=dtype))
	# Set up our output map
	if omap is None: omap_flat = enmap.zeros((ncomp,)+shape[-2:], wcs, dtype)
	else:            omap_flat = omap.preflat
	assert omap_flat.dtype == dtype, "omap.dtype must be np.float32"
	assert omap_flat.shape == (ncomp,)+shape[-2:], "omap must be [...,ny,nx], where [ny,nx] agrees with shape, and ... agrees with amps"
	# Whew! Actually do the work
	times = srcsim.sim_objects(omap_flat, obj_decs, obj_ras, obj_ys, obj_xs, amps_flat, profile, prof_ids, posmap, vmin, rmax=rmax, separable=separable, transpose=transpose, prof_equi=prof_equi, return_times=True)[1]
	omap = omap_flat.reshape(pre+shape[-2:])
	# NB! Since we're not padding, this fourier operation will have problems at the edges
	if pixwin: omap = enmap.apply_window(omap, order=pixwin_order)
	return (omap, times) if return_times else omap

def is_equi(r):
	"""Estimate whether the values r[:] = arange(n)*delta, allowing for
	fast index calculations. This is just a heuristic, but it is hopefully
	reliable enough."""
	return len(r) > 1 and r[0] == 0 and np.allclose(r[-1],(len(r)-1)*r[1])

def radial_sum(map, poss, bins, oprofs=None, separable="auto",
		prof_equi="auto", cache=None, return_times=False):
	"""Sum the signal in map into radial bins around a set of objects,
	returning one radial sum-profile per object.
	Arguments:
	* map: The map to read data from. [...,ny,nx]
	* poss: The positions of the objects. [{dec,ra},nobj] in radians.
	* bins: The bin edges. [nbin+1]. Faster if equi-spaced with first at 0

	Optional arguments:
	* oprofs: [obj,...,nbin] array to write result to. MUST BE float32 AND C CONTIGUOUS
	* separable: Whether the coordinate system's coordinate axes are indpendent,
	  such that one only needs to know y in order to calculate dec, and x to
	  calculate ra. This allows for much faster calculation of the pixel
	  coordinates. Default "auto": True for cylindrical coordinates, False otherwise.
	* cache: Dictionary to use for caching pixel coordinates. Can be useful
	  if you're doing repeated simulations on the same geometry with non-separable
	  geometry, to avoid having to recalculate the pixel coordinates all the time.

	Returns the resulting profiles. If oprof was specified, then the same object will
	be returned (after being updated of course)."""
	dtype = np.float32 # C extension only supports this dtype
	if separable == "auto": separable = wcsutils.is_separable(map.wcs)
	# Object positions
	obj_decs = np.asanyarray(poss[0], dtype=dtype, order="C")
	obj_ras  = np.asanyarray(poss[1], dtype=dtype, order="C")
	obj_ys, obj_xs = utils.nint(map.sky2pix(poss)).astype(np.int32)
	assert obj_decs.ndim == 1 and obj_ras.ndim == 1, "poss must be [{dec,ra},nobj]"
	nobj     = len(obj_decs)
	# map and number of components
	pre      = map.shape[:-2]
	map_flat = map.preflat
	ncomp    = len(map_flat)
	# bins
	bins     = np.asarray(bins, dtype=dtype)
	nbin     = len(bins)-1
	prof_equi = is_equi(bins) if prof_equi == "auto" else prof_equi
	# Set up the pixel coordinates
	if separable: posmap = utils.cache_get(cache, "posmap", lambda: map.posaxes(dtype=dtype))
	else:         posmap = utils.cache_get(cache, "posmap", lambda: map.posmap (dtype=dtype))
	# Set up our output
	if oprofs is None: oprofs = np.zeros((nobj,)+pre+(nbin,),dtype)
	oprofs_flat   = oprofs.reshape(nobj,ncomp,nbin)
	assert oprofs_flat.dtype == dtype, "oprofs.dtype must be np.float32"
	# Whew! Actually do the work
	times = srcsim.radial_sum(map_flat, obj_decs, obj_ras, obj_ys, obj_xs, bins, posmap, profs=oprofs_flat, separable=separable, prof_equi=prof_equi, return_times=True)[1]
	return (oprofs, times) if return_times else oprofs

def radial_bin(map, poss, bins, weights=None, separable="auto",
		prof_equi="auto", cache=None, return_times=False):
	"""Average the signal in map into radial bins for a set of objects, returning
	a radial profile for each object.
	Arguments:
	* map: The map to read data from. [...,ny,nx]
	* poss: The positions of the objects. [{dec,ra},nobj] in radians.
	* bins: The bin edges. [nbin+1]. Faster if equi-spaced with first at 0

	Optional arguments:
	* oprofs: [obj,...,nbin] array to write result to. MUST BE float32 AND C CONTIGUOUS
	* separable: Whether the coordinate system's coordinate axes are indpendent,
	  such that one only needs to know y in order to calculate dec, and x to
	  calculate ra. This allows for much faster calculation of the pixel
	  coordinates. Default "auto": True for cylindrical coordinates, False otherwise.
	* cache: Dictionary to use for caching pixel coordinates. Can be useful
	  if you're doing repeated simulations on the same geometry with non-separable
	  geometry, to avoid having to recalculate the pixel coordinates all the time.

	Returns the resulting profiles. If oprof was specified, then the same object will
	be returned (after being updated of course)."""
	if weights is not None: map = map*weights
	profs, times1 = radial_sum(map, poss, bins, separable=separable, prof_equi=prof_equi,
			cache=cache, return_times=True)
	if weights is None: weights = enmap.ones(map.shape[-2:], map.wcs, map.dtype)
	div, times2 = radial_sum(weights, poss, bins, separable=separable, prof_equi=prof_equi,
			cache=cache, return_times=True)
	# Add dimensions to div if necessary to make them broadcast.
	# For more complex broadcasting, prepare the shapes of map and weights manually
	div = div.reshape(profs.shape[:1]+(1,)*(profs.ndim-div.ndim)+profs.shape[-1:])
	profs /= div
	times = np.concatenate([times1,times2])
	return (profs, times) if return_times else profs

def sim_srcs(shape, wcs, srcs, beam, omap=None, dtype=None, nsigma=5, rmax=None, vmin=None, smul=1,
		return_padded=False, pixwin=False, pixwin_order=0, op=np.add, wrap="auto", verbose=False, cache=None,
		separable="auto", method="c"):
	"""Backwards compatibility wrapper that exposes the speed of the new sim_objects
	function using the old sim_srcs interface. For most users this should result in a
	transparent speedup of O(10x), but sim_objects does not implement 100% of the sim_srcs
	functionality, so the old python method is also available by specifying method = "python".

	Limitations of the new version:
	* only float32, C-contiguous maps supported
	* smul not supported
	* padding not supported, which impacts objects at the very edge of the map if
	  pixwin is used
	* only add, max and min supported for 'op'

	Unlike sim_srcs, sim_objects supports simulating objects with multiple
	different profiles at once, but this functionality isn't available through
	the sim_srcs interface.

	I recommend using sim_objects directly instead of relying on this wrapper in most cases.
	"""
	if method in ["c", "C"]:
		assert dtype is None or dtype == np.float32, "method 'c' only supports float32. Use method='python' if you need others"
		assert smul == 1, "method 'c' does not support smul != 1. Use method='python' if you need this"
		if vmin is None: vmin = np.exp(-0.5*nsigma**2)
		if   op == np.add or op == np.ndarray.__iadd__ or op == "add": op_ = "add"
		elif op == np.max or op == "max": op_ = "max"
		elif op == np.min or op == "min": op_ = "min"
		else: raise ValueError("method 'c' only supports op add, max or min. Use method='python' if you need more")
		if return_padded: raise ValueError("method 'c' does not support return_padded. Use method=='python' if you need this")
		ncomp = np.prod(shape[:-2], dtype=int)
		nobj  = len(srcs)
		poss  = srcs.T[:2]
		amps  = np.zeros((ncomp,nobj), np.float32)
		amps[:srcs.shape[1]-2] = srcs.T[2:2+ncomp]
		amps  = amps.reshape(shape[:-2]+(nobj,))
		# Handle beam = float, in which case a gaussian beam is made
		beam  = expand_beam(beam, nsigma, rmax)
		return sim_objects(shape, wcs, poss, amps, beam, omap=omap, vmin=vmin, rmax=rmax, op=op, pixwin=pixwin, pixwin_order=0, separable=separable, cache=cache)
	elif method == "python":
		return sim_srcs_python(shape, wcs, srcs, beam, omap=omap, dtype=dtype, nsigma=nsigma, rmax=rmax, smul=smul, return_padded=return_padded, pixwin=pixwin, pixwin_order=pixwin_order, op=op, wrap=wrap, verbose=verbose, cache=cache, separable=separable)

def sim_srcs_python(shape, wcs, srcs, beam, omap=None, dtype=None, nsigma=5, rmax=None, smul=1,
		return_padded=False, pixwin=False, pixwin_order=0, op=np.add, wrap="auto", verbose=False, cache=None,
		separable="auto"):
	"""Simulate a point source map in the geometry given by shape, wcs
	for the given srcs[nsrc,{dec,ra,T...}], using the beam[{r,val},npoint],
	which must be equispaced. If omap is specified, the sources will be
	added to it in place. All angles are in radians. The beam is only evaluated up to
	the point where it reaches exp(-0.5*nsigma**2) unless rmax is specified, in which
	case this gives the maximum radius. smul gives a factor to multiply the resulting
	source model by. This is mostly useful in conction with omap.

	The source simulation is sped up by using a source lookup grid.
	"""
	if separable == "auto": separable = wcsutils.is_separable(wcs)
	if omap is None: omap = enmap.zeros(shape, wcs, dtype)
	ishape = omap.shape
	omap   = omap.preflat
	ncomp  = omap.shape[0]
	srcs   = np.asarray(srcs)
	# Set up wrapping
	if utils.streq(wrap, "auto"):
		wrap = [0, utils.nint(360./wcs.wcs.cdelt[0])]
	# In keeping with the rest of the functions here, srcs is [nsrc,{dec,ra,T,Q,U}].
	# The beam parameters are ignored - the beam argument is used instead
	amps = srcs[:,2:2+ncomp]
	poss = srcs[:,:2].copy()
	# Rewind positions to let us use flat-sky approximation for distance calculations
	ref  = np.mean(enmap.box(shape, wcs, corner=False)[:,1])
	poss[:,1] = utils.rewind(poss[:,1], ref)
	beam = expand_beam(beam, nsigma, rmax)
	if not rmax: rmax = nsigma2rmax(beam, nsigma)
	# Pad our map by rmax, so we get the contribution from sources
	# just ourside our area. We will later split our map into cells of size cres. Let's
	# adjust the padding so we have a whole number of cells
	minshape = np.min(omap[...,5:-5:10,5:-5:10].pixshapemap()/10,(-2,-1))
	cres = np.maximum(1,utils.nint(rmax/minshape))
	epix = cres-(omap.shape[-2:]+2*cres)%cres
	padding = [cres,cres+epix]
	wmap, wslice  = enmap.pad(omap, padding, return_slice=True)
	# Overall we will have this many grid cells
	cshape = wmap.shape[-2:]//cres
	# Find out which sources matter for which cells
	srcpix = wmap.sky2pix(poss.T).T
	pixbox= np.array([[0,0],wmap.shape[-2:]],int)
	nhit, cell_srcs = build_src_cells(pixbox, srcpix, cres, wrap=wrap)
	# Optionally cache the posmap
	if cache is None or cache[0] is None: posmap = wmap.posmap(separable=separable)
	else: posmap = cache[0]
	if cache is not None: cache[0] = posmap
	model = eval_srcs_loop(posmap, poss, amps, beam, cres, nhit, cell_srcs, dtype=wmap.dtype, op=op, verbose=verbose)
	del posmap
	if pixwin: model = enmap.apply_window(model, order=pixwin_order)
	# Update our work map, through our view
	if smul != 1: model *= smul
	wmap   = op(wmap, model, wmap)
	if not return_padded:
		# Copy out
		omap[:] = wmap[wslice]
		# Restore shape
		omap = omap.reshape(ishape)
		return omap
	else:
		return wmap.reshape(ishape[:-2]+wmap.shape[-2:]), wslice

def eval_srcs_loop(posmap, poss, amps, beam, cres, nhit, cell_srcs, dtype=np.float64, op=np.add, verbose=False):
	# Loop through each cell
	ncy, ncx = nhit.shape
	model = enmap.zeros(amps.shape[-1:]+posmap.shape[-2:], posmap.wcs, dtype)
	for cy in range(ncy):
		for cx in range(ncx):
			nsrc = nhit[cy,cx]
			if verbose and nsrc > 0: print("map cell %5d/%d with %5d srcs" % (cy*ncx+cx+1, ncy*ncx, nsrc))
			if nsrc == 0: continue
			srcs  = cell_srcs[cy,cx,:nsrc]
			y1,y2 = (cy+0)*cres[0], (cy+1)*cres[0]
			x1,x2 = (cx+0)*cres[1], (cx+1)*cres[1]
			pixpos = posmap[:,y1:y2,x1:x2]
			srcpos = poss[srcs].T # [2,nsrc]
			srcamp = amps[srcs].T # [ncomp,nsrc]
			r      = utils.angdist(pixpos[::-1,None,:,:],srcpos[::-1,:,None,None])
			bpix   = (r - beam[0,0])/(beam[0,1]-beam[0,0])
			# Evaluate the beam at these locations
			bval   = utils.interpol(beam[1], bpix[None], mode="constant", order=1, mask_nan=False) # [nsrc,ry,rx]
			cmodel = srcamp[:,:,None,None]*bval
			cmodel = op.reduce(cmodel,-3)
			op(model[:,y1:y2,x1:x2], cmodel, model[:,y1:y2,x1:x2])
	return model

def sim_srcs_dist_transform(shape, wcs, srcs, beam, omap=None, dtype=None, nsigma=4, rmax=None, smul=1,
		pixwin=False, ignore_outside=False, op=np.add, verbose=False):
	"""Simulate a point source map in the geometry given by shape, wcs
	for the given srcs[nsrc,{dec,ra,T...}], using the beam[{r,val},npoint],
	which must be equispaced. Unlike sim_srcs, overalpping point sources are not supported.
	If omap is specified, the sources will be
	added to it in place. All angles are in radians. The beam is only evaluated up to
	the point where it reaches exp(-0.5*nsigma**2) unless rmax is specified, in which
	case this gives the maximum radius. smul gives a factor to multiply the resulting
	source model by. This is mostly useful in conction with omap.
	"""
	if omap is None: omap = enmap.zeros(shape, wcs, dtype)
	ishape = omap.shape
	omap   = omap.preflat
	ncomp  = omap.shape[0]
	# In keeping with the rest of the functions here, srcs is [nsrc,{dec,ra,T,Q,U}].
	# The beam parameters are ignored - the beam argument is used instead
	amps = srcs[:,2:2+ncomp]
	poss = srcs[:,:2].copy()
	br, bv = expand_beam(beam, nsigma, rmax).copy()
	rmax = nsigma2rmax(beam, nsigma)
	if ignore_outside:
		pixs = enmap.sky2pix(shape, wcs, poss.T[:2])
		pmax = 2*rmax/enmap.pixsize(shape, wcs)**0.5
		bad  = np.any(pixs <  -pmax, 0) | np.any(pixs > np.array(shape[-2:])[:,None]+pmax, 0)
		amps, poss = amps[~bad], poss[~bad]
	r, domains = enmap.distance_from(shape, wcs, poss.T[:2], domains=True, rmax=rmax)
	np.clip(domains, 0, amps.shape[0]-1, domains)
	for i in range(ncomp):
		op(omap[i], np.interp(r, br, bv, right=0) * amps[domains,i], omap[i])
	omap = omap.reshape(ishape)
	return omap

def expand_beam(beam, nsigma=5, rmax=None, nper=400):
	beam = np.asarray(beam)
	if beam.ndim == 0:
		# Build gaussian beam
		sigma = beam.reshape(-1)[0]
		if rmax is None: rmax = sigma*nsigma
		r = np.linspace(0,rmax,nsigma*nper)
		return np.array([r,np.exp(-0.5*(r/sigma)**2)])
	elif beam.ndim == 2:
		return beam
	else: raise ValueError

def nsigma2rmax(beam, nsigma):
	return beam[0,np.where(beam[1] >= np.exp(-0.5*nsigma**2))[0][-1]]

def build_src_cells(cbox, srcpos, cres, unwind=False, wrap=None):
	# srcpos is [nsrc,...,{dec,ra}]. Reshape to 3d to keep things simple.
	# will reshape back when returning
	cbox    = np.asarray(cbox)
	srcpos  = np.asarray(srcpos)
	ishape  = srcpos.shape
	srcpos  = srcpos.reshape(ishape[0],-1,ishape[-1])

	cshape  = tuple(np.ceil(((cbox[1]-cbox[0])/cres)).astype(int))
	if unwind:
		# Make the sources' ra compatible with our area
		ref     = np.mean(cbox[:,1],0)
		srcpos[:,...,1] = utils.rewind(srcpos[:,...,1], ref)
	# How big must our cell hit array be?
	nmax = max(1,np.max(build_src_cells_helper(cbox, cshape, cres, srcpos, wrap=wrap)))
	ncell, cells = build_src_cells_helper(cbox, cshape, cres, srcpos, nmax, wrap=wrap)
	# Reshape back to original shape
	ncell = ncell.reshape(ishape[1:-1]+ncell.shape[1:])
	cells = cells.reshape(ishape[1:-1]+cells.shape[1:])
	return ncell, cells

def build_src_cells_helper(cbox, cshape, cres, srcpos, nmax=0, wrap=None):
	# A cell is hit if it overlaps both horizontally and vertically
	# with the point source +- cres
	nsrc, nmid = srcpos.shape[:2]
	# ncell is [:,ncy,ncx]
	ncell = np.zeros((nmid,)+cshape,np.int32)
	if nmax > 0:
		cells = np.zeros((nmid,)+cshape+(nmax,),np.int32)
	c0 = cbox[0]; inv_dc = cshape/(cbox[1]-cbox[0]).astype(float)
	# Set up wrapping. The woff variable will contain the set of coordinate offsets we will try
	if wrap is None: wrap = [0,0]
	woffs = []
	for i, w in enumerate(wrap):
		if w == 0: woffs.append([0])
		else: woffs.append([-w,0,+w])
	for si in range(nsrc):
		for mi in range(nmid):
			pos = srcpos[si,mi]
			for woffy in woffs[0]:
				for woffx in woffs[1]:
					wpos = pos[:2] + np.array([woffy,woffx])
					i1   = (wpos[:2]-cres-c0)*inv_dc
					i2   = (wpos[:2]+cres-c0)*inv_dc+1 # +1 because this is a half-open interval
					# Don't try to update out of bounds
					i1 = np.maximum(i1.astype(int), 0)
					i2 = np.minimum(i2.astype(int), np.array(cshape))
					# Skip sources that don't hit our area at all
					if np.any(i1 >= cshape) or np.any(i2 < 0): continue
					for cy in range(i1[0],i2[0]):
						for cx in range(i1[1],i2[1]):
							if nmax > 0:
								cells[mi,cy,cx,ncell[mi,cy,cx]] = si
							ncell[mi,cy,cx] += 1
	if nmax > 0: return ncell, cells
	else: return ncell

def cellify(map, res):
	"""Given a map [...,ny,nx] and a cell resolution [ry,rx], return map
	reshaped into a cell grid [...,ncelly,ncellx,ry,rx]. The map will be
	truncated if necessary"""
	res    = np.array(res,int)
	cshape = map.shape[-2:]//res
	omap   = map[...,:cshape[0]*res[0],:cshape[1]*res[1]]
	omap   = omap.reshape(omap.shape[:-2]+(cshape[0],res[0],cshape[1],res[1]))
	omap   = utils.moveaxis(omap, -3, -2)
	return omap

def uncellify(cmap):
	omap = utils.moveaxis(cmap, -2, -3)
	omap = omap.reshape(omap.shape[:-4]+(omap.shape[-4]*omap.shape[-3],omap.shape[-2]*omap.shape[-1]))
	return omap

#### Cross-matching ####

def crossmatch(srcs1, srcs2, tol=1*utils.arcmin, safety=4):
	"""Cross-match two source catalogs based on position. Each
	source in one catalog is associated with the closest source
	in the other catalog, as long as the distance between them is
	less than the tolerance. The catalogs must be [:,{ra,dec,...}]
	in radians. Returns [nmatch,2], with the last index giving
	the index in the first and second catalog for each match."""
	vec1 = utils.ang2rect(srcs1[:,:2], axis=1)
	vec2 = utils.ang2rect(srcs2[:,:2], axis=1)
	tree1 = spatial.cKDTree(vec1)
	tree2 = spatial.cKDTree(vec2)
	groups = tree1.query_ball_tree(tree2, tol*safety)
	matches = []
	for gi, group in enumerate(groups):
		if len(group) == 0: continue
		# Get the true distance to each member in the group
		group = np.array(group)
		dists = utils.vec_angdist(vec1[gi,None,:], vec2[group,:], axis=1)
		best  = np.argmin(dists)
		if dists[best] > tol: continue
		matches.append([gi, group[best]])
	matches = np.array(matches)
	return matches

#### Source parameter I/O ####

# These functions are messy and should be cleaned up before general consumption
# Be warned if you choose to use them - they might be removed or changed in future
# releases.

def read(fname, format="auto"):
	if format == "auto": formats = ["dory_fits","dory_txt","fits","nemo","simple"]
	else:                formats = [format]
	for f in formats:
		try:
			if   f == "dory_fits":return read_dory_fits(fname)
			elif f == "dory_txt": return read_dory_txt(fname)
			elif f == "fits":     return read_fits(fname)
			elif f == "nemo":     return read_nemo(fname)
			elif f == "simple":   return read_simple(fname)
			else: raise ValueError("Unrecognized point source format '%s' for file '%s'" % (f, fname))
		except (ValueError, IOError, OSError) as e: pass
	raise IOError("Unable to read point source file '%s' with format '%s'" % (fname, f))

def read_nemo(fname):
	"""Reads the nemo ascii catalog format, and returns it as a recarray."""
	idtype = [("name","2S64"),("ra","d"),("dec","d"),("snr","d"),("npix","i"),("detfrac","d"),("template","S32"),("glat","d"),("I","d"), ("dI","d")]
	try: icat = np.loadtxt(fname, dtype=idtype)
	except (ValueError, IndexError) as e:
		idtype = [("name","2S64"),("ra","d"),("dec","d"),("snr","d"),("npix","i"),("template","S32"),("glat","d"),("I","d"), ("dI","d")]
		try: icat = np.loadtxt(fname, dtype=idtype)
		except (ValueError, IndexError) as e:
			raise IOError(e.args[0])
	odtype = [("name","S64"),("ra","d"),("dec","d"),("snr","d"),("I","d"),("dI","d"),("npix","i"),("template","S32"),("glat","d")]
	ocat = np.zeros(len(icat), odtype).view(np.recarray)
	ocat.name = np.char.add(*icat["name"].T)
	for t in odtype[1:]: ocat[t[0]] = icat[t[0]]
	return ocat

def read_simple(fname):
	try:
		cat = np.loadtxt(fname, dtype=[("ra","d"),("dec","d"),("I","d"),("dI","d")], usecols=range(4), ndmin=1).view(np.recarray)
	except ValueError:
		try:
			cat = np.loadtxt(fname, dtype=[("ra","d"),("dec","d"),("I","d")], usecols=range(3), ndmin=1).view(np.recarray)
		except ValueError as e:
			raise IOError(e.args[0])
	cat.ra  *= utils.degree
	cat.dec *= utils.degree
	return cat

def read_dory_fits(fname, hdu=1):
	d = fits.open(fname)[hdu].data
	ocat = np.zeros(len(d), dtype=[("ra","d"),("dec","d"),("I","d"),("Q","d"),("U","d")]).view(np.recarray)
	ocat.ra  = d.ra  * utils.degree
	ocat.dec = d.dec * utils.degree
	ocat.I, ocat.Q, ocat.U = d.amp.T*1e3
	return ocat

def read_dory_txt(fname):
	try:
		d = np.loadtxt(fname, usecols=[0,1,3,5,7], dtype=[("ra","d"),("dec","d"),("I","d"),("Q","d"),("U","d")]).view(np.recarray).reshape(-1)
		d.I   *= 1e3; d.Q *= 1e3; d.U *= 1e3
		d.ra  *= utils.degree
		d.dec *= utils.degree
		return d
	except (ValueError, IndexError) as e:
		raise IOError(e.args[0])

def read_fits(fname, hdu=1, fix=True):
	d = fits.open(fname)[hdu].data
	if fix:
		d = translate_dtype_keys(d, {"RADeg":"ra","decDeg":"dec","deltaT_c":"I","err_deltaT_c":"dI"})
	return d.view(np.recarray)

# Sauron catalog formats. I really need to come up with some once-and-for-all catalog format!
def format_sauron(cat):
	nfield, ncomp = cat.flux.shape[-2:]
	names  = "TQU"
	header = "#%8s %8s %9s" % ("ra", "dec", "snr_T")
	for i in range(1,ncomp): header += " %8s" % ("snr_"+names[i])
	for i in range(ncomp):   header += " %8s %7s" % ("ftot_"+names[i], "dftot_"+names[i])
	for i in range(nfield):
		for j in range(ncomp):
			header += " %8s %7s" % ("flux_"+names[j]+"%d"%(i+1), "dflux_"+names[j]+"%d"%(i+1))
	header += " %2s" % "ca"
	for i in range(nfield): header += " %7s" % ("cont_%d" % (i+1))
	header += "\n"
	res = ""
	for i in range(len(cat)):
		res += "%9.4f %8.4f" % (cat.ra[i]/utils.degree, cat.dec[i]/utils.degree)
		snr  = cat.snr[i].reshape(-1)
		res += " %9.2f" % snr[0] + " %7.2f"*(len(snr)-1) % tuple(snr[1:])
		flux = cat. flux_tot[i].reshape(-1)
		dflux= cat.dflux_tot[i].reshape(-1)
		for j in range(len(flux)):
			res += "  %8.2f %7.2f" % (flux[j], dflux[j])
		flux = cat. flux[i].reshape(-1)
		dflux= cat.dflux[i].reshape(-1)
		for j in range(len(flux)):
			res += "  %8.2f %7.2f" % (flux[j], dflux[j])
		try: res += " %2d" % (cat.case[i])
		except (KeyError, AttributeError): pass
		try:
			for j in range(len(cat.contam[i])):
				res += " %7.2f" % (cat.contam[i,j])
		except (KeyError, AttributeError): pass
		res += "\n"
	return header + res

def write_sauron(ofile, cat):
	if ofile.endswith(".fits"): write_sauron_fits(ofile, cat)
	else: write_sauron_txt (ofile, cat)

def read_sauron(ifile):
	if ifile.endswith(".fits"): return read_sauron_fits(ifile)
	else: return read_sauron_txt(ifile)

def write_sauron_fits(ofile, cat):
	from astropy.io import fits
	ocat = cat.copy()
	for field in ["ra","dec"]: ocat[field] /= utils.degree # angles in degrees
	hdu = fits.hdu.table.BinTableHDU(ocat)
	hdu.writeto(ofile, overwrite=True)

def read_sauron_fits(fname):
	from astropy.io import fits
	hdu = fits.open(fname)[1]
	cat = np.asarray(hdu.data).view(np.recarray)
	for field in ["ra","dec"]: cat[field] *= utils.degree # deg -> rad
	return cat

def write_sauron_txt(ofile, cat):
	with open(ofile, "w") as ofile:
		ofile.write(format_sauron(cat))

def read_sauron_txt(ifile, ncomp=3):
	raw   = np.loadtxt(ifile, ndmin=2)
	nrow, ncol = raw.shape
	nfreq = (ncol-2-ncomp-1)//(2*ncomp+1)
	cat_dtype  = [("ra", "d"), ("dec", "d"), ("snr", "d", (ncomp,)), ("flux_tot", "d", (ncomp,)),
			("dflux_tot", "d", (ncomp,)), ("flux", "d", (nfreq,ncomp)), ("dflux", "d", (nfreq,ncomp)),
			("case", "i"), ("contam", "d", (nfreq,))]
	ocat  = np.zeros(nrow, cat_dtype).view(np.recarray)
	ocat.ra, ocat.dec, raw = raw[:,0]*utils.degree, raw[:,1]*utils.degree, raw[:,2:]
	ocat.snr,          raw = raw[:,:ncomp], raw[:,ncomp:]
	ocat.flux_tot, ocat.dflux_tot, raw = raw[:,0:2*ncomp:2], raw[:,1:2*ncomp:2], raw[:,2*ncomp:]
	ocat.flux,     ocat.dflux,     raw = raw[:,0:2*ncomp*nfreq:2].reshape(-1,nfreq,ncomp), raw[:,1:2*ncomp*nfreq:2].reshape(-1,nfreq,ncomp), raw[:,2*ncomp*nfreq:]
	ocat.case,         raw = raw[:,0], raw[:,1:]
	ocat.contam            = raw[:,:nfreq]
	return ocat

def translate_dtype_keys(d, translation):
	descr = [(name if name not in translation else translation[name], char) for (name, char) in d.dtype.descr]
	return np.asarray(d, descr)

def src2param(srcs):
	"""Translate recarray srcs into the source fromat used for tod-level point source
	operations."""
	params = np.zeros(srcs.shape + (8,))
	params[:,0] = srcs.dec * utils.degree # yes, dec first
	params[:,1] = srcs.ra  * utils.degree
	params[:,2] = srcs.I
	if "Q" in srcs.dtype.names: params[:,3] = srcs.Q
	if "U" in srcs.dtype.names: params[:,4] = srcs.U
	# These are not used
	params[:,5] = 1 # x-scaling
	params[:,6] = 1 # y-scaling
	params[:,7] = 0 # angle
	return params



import cython, numpy as np
from scipy import ndimage
cimport numpy as np
from libc.stdlib cimport free
# I had to call this distances_wrap instead of just distances because
# .pdx files with the same name get automatically imported into the current
# namespace, which means I would have to give stupid names to the functions here
cimport distances_wrap as c
from distances_wrap cimport inum, uint8_t

__version__ = 1.0

def distance_from_points_simple(posmap, points, omap=None, odomains=None, domains=False):
	"""distance_from_points(posmap, points, omap=None, odomains=None, domains=False)

	Given a posmap[{dec,ra},ny,nx] and a set of points[{dec,ra},npoint], computes the
	angular distance map from every pixel [ny,nx] to the nearest point. If domains==True,
	then a [ny,nx] map of the index of the nearest point is also returned. New arrays
	will be created for the output unless omap and/or odomains are specified, in which
	case they will be overwritten."""
	# Check that our inputs make sense
	posmap = np.asanyarray(posmap).astype(float, order="C", copy=False)
	points = np.asanyarray(points).astype(float, order="C", copy=False)
	assert posmap.ndim == 3 and len(posmap) == 2, "posmap must be [{dec,ra},ny,nx]"
	assert points.ndim == 2 and len(points) == 2, "points must be [{dec,ra},npoint]"
	if omap is None: omap = np.empty_like(posmap[0], dtype=np.float64)
	assert omap.ndim == 2 and omap.shape[-2:] == posmap.shape[-2:] and omap.dtype==np.float64, "omap must be [ny,nx] float64"
	if domains:
		if odomains is None: odomains = np.empty_like(posmap[0], dtype=np.int32)
		assert odomains.ndim == 2 and odomains.shape[-2:] == posmap.shape[-2:] and odomains.dtype==np.int32, "odomains must be [ny,nx] int32"
	# Prepare to call C
	cdef inum npix   = omap.size
	cdef inum npoint = points.shape[1]
	cdef double[::1] posmap_ = posmap.reshape(-1)
	cdef double[::1] points_ = points.reshape(-1)
	cdef double[::1] omap_   = omap.reshape(-1)
	cdef int[::1]    odomains_
	if not domains:
		c.distance_from_points_simple(npix, &posmap_[0], npoint, &points_[0], &omap_[0], NULL)
		return omap
	else:
		odomains_ = odomains.reshape(-1)
		c.distance_from_points_simple(npix, &posmap_[0], npoint, &points_[0], &omap_[0], &odomains_[0])
		return omap, odomains

def distance_from_points_simple_separable(ypos, xpos, points, omap=None, odomains=None, domains=False):
	"""distance_from_points_simple_separable(ypos, xpos, points, omap=None, odomains=None, domains=False)

	Like distance_from_points, but optimized for the case where the coordinate system
	is separable, as is typically the case for cylindrical projections. Instead of a full
	posmap[{dec,ra},ny,nx] it takes ypos[ny] which gives the dec of each point along the y axis
	and xpos[nx] which gives the ra of each point along the x axis. The main advantage of this
	is that one can avoid the somewhat heavy computation of the full posmap."""
	# Check that our inputs make sense
	ypos   = np.asanyarray(ypos).astype(float, order="C", copy=False)
	xpos   = np.asanyarray(xpos).astype(float, order="C", copy=False)
	points = np.asanyarray(points).astype(float, order="C", copy=False)
	assert ypos.ndim == 1, "ypos must be [ny]"
	assert xpos.ndim == 1, "xpos must be [nx]"
	assert points.ndim == 2 and len(points) == 2, "points must be [{dec,ra},npoint]"
	if omap is None: omap = np.empty_like(ypos, dtype=np.float64, shape=(len(ypos),len(xpos)))
	assert omap.ndim == 2 and omap.shape[-2:] == (len(ypos),len(xpos)) and omap.dtype==np.float64, "omap must be [ny,nx] float64"
	if domains:
		if odomains is None: odomains = np.empty_like(ypos, dtype=np.int32, shape=(len(ypos),len(xpos)))
		assert odomains.ndim == 2 and odomains.shape[-2:] == (len(ypos),len(xpos)) and odomains.dtype==np.int32, "odomains must be [ny,nx] int32"
	# Prepare to call C
	cdef inum ny = len(ypos)
	cdef inum nx = len(xpos)
	cdef inum npoint = points.shape[1]
	cdef double[::1] ypos_   = ypos.reshape(-1)
	cdef double[::1] xpos_   = xpos.reshape(-1)
	cdef double[::1] points_ = points.reshape(-1)
	cdef double[::1] omap_   = omap.reshape(-1)
	cdef int[::1]    odomains_
	if not domains:
		c.distance_from_points_simple_separable(ny, nx, &ypos_[0], &xpos_[0], npoint, &points_[0], &omap_[0], NULL)
		return omap
	else:
		odomains_ = odomains.reshape(-1)
		c.distance_from_points_simple_separable(ny, nx, &ypos_[0], &xpos_[0], npoint, &points_[0], &omap_[0], &odomains_[0])
		return omap, odomains

def fix_point_pix(posmap, point_pos, point_pix):
	"""fix_point_pix(posmap, point_pos, point_pix)

	Return a new point_pix where out of bounds points have been replaced by the closest point
	on the boundary. This uses the simple method for the pixels on the boundary. The number of pixels
	on the boundary are much lower than the pixels in the main part of the image, which partially
	makes up for the slowness of the simple method, but you can still expect slowness if too many points
	are outside the image."""
	ypos, xpos = posmap
	# Get the bad points
	bad = np.where(np.any(point_pix < 0,0) | np.any(point_pix >= np.array(ypos.shape[-2:])[:,None],0))[0]
	# If there aren't any, we can just return right away
	nbad = len(bad)
	if nbad == 0: return point_pix
	# Otherwise, run the simple method on each boundary. The simple method is slow, but the boundary is
	# has much fewer pixels.
	point_bad = point_pos[:,bad]
	pos_edges = np.array([np.concatenate([p[:,0],p[:,-1],p[0,:],p[-1,:]],-1) for p in [ypos,xpos]])
	dist, dom = distance_from_points_simple(pos_edges[:,None,:], point_bad, domains=True)
	dist, dom = dist[0], dom[0]
	# Find the minimum position for each point
	minpos1d = np.array(ndimage.minimum_position(dist, dom+1, np.arange(nbad)+1)).reshape(-1)
	# Turn the 1d minpos into a 2d pixel position
	minpos   = _unwrap_minpos(minpos1d, ypos.shape[0], ypos.shape[1])
	# Copy these into the output point_pix
	opoint_pix = point_pix.copy()
	opoint_pix[:,bad] = minpos
	return opoint_pix

def fix_point_pix_separable(ypos, xpos, point_pos, point_pix):
	"""fix_point_pix_separable(ypos, xpos, point_pos, point_pix)

	Return a new point_pix where out of bounds points have been replaced by the closest point
	on the boundary. This uses the simple method for the pixels on the boundary. The number of pixels
	on the boundary are much lower than the pixels in the main part of the image, which partially
	makes up for the slowness of the simple method, but you can still expect slowness if too many points
	are outside the image."""
	# Get the bad points
	bad = np.where(np.any(point_pix < 0,0) | (point_pix[0] >= ypos.size) | (point_pix[1] >= xpos.size))[0]
	# If there aren't any, we can just return right away
	nbad = len(bad)
	if nbad == 0: return point_pix
	# Otherwise, run the simple method on each boundary. The simple method is slow, but the boundary is
	# has much fewer pixels.
	point_bad = point_pos[:,bad]
	pos_edges = np.array([
		np.concatenate([ypos, ypos, np.full(xpos.size, ypos[0]), np.full(xpos.size, ypos[-1])]),
		np.concatenate([np.full(ypos.size, xpos[0]), np.full(ypos.size, xpos[-1]), xpos, xpos])])
	dist, dom = distance_from_points_simple(pos_edges[:,None,:], point_bad, domains=True)
	dist, dom = dist[0], dom[0]
	# Find the minimum position for each point
	minpos1d = np.array(ndimage.minimum_position(dist, dom+1, np.arange(nbad)+1)).reshape(-1)
	# Turn the 1d minpos into a 2d pixel position
	minpos   = _unwrap_minpos(minpos1d, ypos.size, xpos.size)
	# Copy these into the output point_pix
	opoint_pix = point_pix.copy()
	opoint_pix[:,bad] = minpos
	return opoint_pix

def _unwrap_minpos(minpos1d, ny, nx):
	minpos   = np.zeros([2,len(minpos1d)], minpos1d.dtype)
	mask = minpos1d < ny;
	minpos[0,mask], minpos[1,mask] = minpos1d[mask], 0
	mask = (minpos1d >= ny) & (minpos1d < 2*ny)
	minpos[0,mask], minpos[1,mask] = minpos1d[mask]-ny, nx-1
	mask = (minpos1d >= 2*ny) & (minpos1d < 2*ny+nx)
	minpos[0,mask], minpos[1,mask] = 0, minpos1d[mask]-2*ny
	mask = minpos1d >= 2*ny+nx
	minpos[0,mask], minpos[1,mask] = ny-1, minpos1d[mask]-(2*ny+nx)
	return minpos

def distance_from_points_bubble(posmap, point_pos, point_pix, rmax=None, omap=None, odomains=None, domains=False):
	"""distance_from_points_buble(posmap, point_pos, point_pix, rmax=None, omap=None, odomains=None, domains=False)"""
	# Check that our inputs make sense
	posmap    = np.asanyarray(posmap).astype(float, order="C", copy=False)
	point_pos = np.asanyarray(point_pos).astype(float,    order="C", copy=False)
	point_pix = np.asanyarray(point_pix).astype(np.int32, order="C", copy=False)
	assert posmap.ndim == 3 and len(posmap) == 2, "posmap must be [{dec,ra},ny,nx]"
	assert point_pos.ndim == 2 and len(point_pos) == 2, "point_pos must be [{dec,ra},npoint]"
	assert point_pix.ndim == 2 and len(point_pix) == 2, "point_pix must be [{y,x},npoint]"
	if omap is None: omap = np.empty_like(posmap[0], dtype=np.float64)
	assert omap.ndim == 2 and omap.shape == posmap.shape[-2:] and omap.dtype==np.float64, "omap must be [ny,nx] float64"
	if odomains is None: odomains = np.empty_like(posmap[0], dtype=np.int32)
	assert odomains.ndim == 2 and odomains.shape == posmap.shape[-2:] and odomains.dtype==np.int32, "odomains must be [ny,nx] int32"
	# Make point_pix safe
	point_pix = fix_point_pix(posmap, point_pos, point_pix)
	# Prepare to call C
	cdef int ny = posmap.shape[1]
	cdef int nx = posmap.shape[2]
	cdef inum npoint = point_pos.shape[1]
	cdef double      rmax_      = (0 if rmax is None else rmax)
	cdef double[::1] posmap_    = posmap.reshape(-1)
	cdef double[::1] point_pos_ = point_pos.reshape(-1)
	cdef int[::1]    point_pix_ = point_pix.reshape(-1)
	cdef double[::1] omap_   = omap.reshape(-1)
	cdef int[::1]    odomains_ = odomains.reshape(-1)
	c.distance_from_points_bubble(ny, nx, &posmap_[0], npoint, &point_pos_[0], &point_pix_[0], rmax_, &omap_[0], &odomains_[0])
	if domains: return omap, odomains
	else:       return omap

def distance_from_points_bubble_separable(ypos, xpos, point_pos, point_pix, rmax=None, omap=None, odomains=None, domains=False):
	"""distance_from_points_bubble_separable(ypos, xpos, point_pos, point_pix, rmax=None, omap=None, odomains=None, domains=False)

	Like distance_from_points_bubble, but optimized for the case where the coordinate system
	is separable, as is typically the case for cylindrical projections. Instead of a full
	posmap[{dec,ra},ny,nx] it takes ypos[ny] which gives the dec of each point along the y axis
	and xpos[nx] which gives the ra of each point along the x axis. The main advantage of this
	is that one can avoid the somewhat heavy computation of the full posmap. If rmax is specified,
	the calculation will stop at the distance rmax. Beyond that distance domain will be set to -1 and
	distance will be set to rmax. This can be used to speed up the calculation if one only cares about
	distances up to a certain point."""
	# Check that our inputs make sense
	ypos   = np.asanyarray(ypos).astype(float, order="C", copy=False)
	xpos   = np.asanyarray(xpos).astype(float, order="C", copy=False)
	point_pos = np.asanyarray(point_pos).astype(float,    order="C", copy=False)
	point_pix = np.asanyarray(point_pix).astype(np.int32, order="C", copy=False)
	assert ypos.ndim == 1, "ypos must be [ny]"
	assert xpos.ndim == 1, "xpos must be [nx]"
	assert point_pos.ndim == 2 and len(point_pos) == 2, "point_pos must be [{dec,ra},npoint]"
	assert point_pix.ndim == 2 and len(point_pix) == 2 and point_pix.shape[1] == point_pos.shape[1], "point_pos must be [{y,x},npoint]"
	if omap is None: omap = np.empty_like(ypos, dtype=np.float64, shape=(len(ypos),len(xpos)))
	assert omap.ndim == 2 and omap.shape[-2:] == (len(ypos),len(xpos)) and omap.dtype==np.float64, "omap must be [ny,nx] float64"
	if odomains is None: odomains = np.empty_like(ypos, dtype=np.int32, shape=(len(ypos),len(xpos)))
	assert odomains.ndim == 2 and odomains.shape[-2:] == (len(ypos),len(xpos)) and odomains.dtype==np.int32, "odomains must be [ny,nx] int32"
	# Make point_pix safe
	point_pix = fix_point_pix_separable(ypos, xpos, point_pos, point_pix)
	# Prepare to call C
	cdef int ny = len(ypos)
	cdef int nx = len(xpos)
	cdef inum npoint = point_pos.shape[1]
	cdef double      rmax_   = (0 if rmax is None else rmax)
	cdef double[::1] ypos_   = ypos.reshape(-1)
	cdef double[::1] xpos_   = xpos.reshape(-1)
	cdef double[::1] point_pos_ = point_pos.reshape(-1)
	cdef int[::1]    point_pix_ = point_pix.reshape(-1)
	cdef double[::1] omap_   = omap.reshape(-1)
	cdef int[::1]    odomains_ = odomains.reshape(-1)
	c.distance_from_points_bubble_separable(ny, nx, &ypos_[0], &xpos_[0], npoint, &point_pos_[0], &point_pix_[0], rmax_, &omap_[0], &odomains_[0])
	if domains: return omap, odomains
	else:       return omap

def distance_from_points_cellgrid(ypos, xpos, point_pos, point_pix, dr=np.inf, rmax=None, omap=None, odomains=None, domains=False, bsize=32):
	"""distance_from_points_cellgrid(ypos, xpos, point_pos, point_pix, rmax=None, omap=None, odomains=None, domains=False)

	dr = np.inf and bsize=32 are based on benchmarking on my laptop. Reducing dr was never
	beneficial and could damage performance significantly if set too low. bsize in the range
	10-50 are OK, but worst-case performance rapidly gets worse after that. The best bsize was
	slightly dependent on the size of the map
	"""
	ypos   = np.asanyarray(ypos).astype(float, order="C", copy=False)
	xpos   = np.asanyarray(xpos).astype(float, order="C", copy=False)
	point_pos = np.asanyarray(point_pos).astype(float,    order="C", copy=False)
	point_pix = np.asanyarray(point_pix).astype(np.int32, order="C", copy=False)
	cdef int ny
	cdef int nx
	cdef int separable
	posmsg = "ypos must be [ny] or [ny,nx] and xpos must be [nx] or [ny,nx]"
	if ypos.ndim == 1:
		assert xpos.ndim == 1, posmsg
		ny, nx= len(ypos), len(xpos)
		separable = 1
	else:
		assert ypos.ndim == 2, posmsg
		assert xpos.shape == ypos.shape, posmsg
		ny, nx = ypos.shape
		separable = 0
	assert point_pos.ndim == 2 and len(point_pos) == 2, "point_pos must be [{dec,ra},npoint]"
	assert point_pix.ndim == 2 and len(point_pix) == 2 and point_pix.shape[1] == point_pos.shape[1], "point_pos must be [{y,x},npoint]"
	if omap is None: omap = np.empty_like(ypos, dtype=np.float64, shape=(ny,nx))
	assert omap.ndim == 2 and omap.shape[-2:] == (ny,nx) and omap.dtype==np.float64, "omap must be [ny,nx] float64"
	if odomains is None: odomains = np.empty_like(ypos, dtype=np.int32, shape=(ny,nx))
	assert odomains.ndim == 2 and odomains.shape[-2:] == (ny,nx) and odomains.dtype==np.int32, "odomains must be [ny,nx] int32"
	bsize = np.zeros(2,np.int32)+bsize
	# Make point_pix safe
	if separable:
		point_pix = fix_point_pix_separable(ypos, xpos, point_pos, point_pix)
	else:
		point_pix = fix_point_pix([ypos, xpos], point_pos, point_pix)
	# Prepare to call C
	cdef inum npoint = point_pos.shape[1]
	cdef double      rmax_   = (0 if rmax is None else rmax)
	cdef double[::1] ypos_   = ypos.reshape(-1)
	cdef double[::1] xpos_   = xpos.reshape(-1)
	cdef double[::1] point_pos_ = point_pos.reshape(-1)
	cdef int[::1]    point_pix_ = point_pix.reshape(-1)
	cdef double[::1] omap_   = omap.reshape(-1)
	cdef int[::1]    odomains_ = odomains.reshape(-1)
	c.distance_from_points_cellgrid(ny, nx, &ypos_[0], &xpos_[0], npoint, &point_pos_[0], &point_pix_[0], bsize[0], bsize[1], rmax_, dr, separable, &omap_[0], &odomains_[0])
	if domains: return omap, odomains
	else:       return omap


def find_edges(mask, flat=False):
	"""find_edges(mask, flat=False)

	Given a 2d numpy array mask[ny,nx], returns an list of indices (y[:],x[:]) for each
	pixel at the edge of a zero region in the mask. If flat==True then it will instead
	return a list of indicees [:] into the flattened mask."""
	# Ensure that we have the right data type and contiguity. Try extra hard to avoid copies
	# when the input is known to have a compatible data type.
	mask = np.asanyarray(mask)
	if mask.dtype in [np.uint8, np.int8, bool]:
		mask = mask.astype(mask.dtype, order="C", copy=False).view(np.uint8)
	else:
		mask = mask.astype(np.uint8, order="C", copy=False)
	assert mask.ndim == 2, "mask must be 2D"
	cdef inum n = 0
	cdef inum * edges_raw = NULL;
	cdef uint8_t[::1] mask_ = mask.reshape(-1)
	cdef inum[::1] view
	n = c.find_edges(mask.shape[0], mask.shape[1], &mask_[0], &edges_raw)
	# Copy into numpy array
	if n > 0:
		view = <inum[:n]> edges_raw;
		edges = np.array(view)
	else:
		edges = np.zeros(0, np.int64)
	free(edges_raw)
	if not flat:
		edges = np.unravel_index(edges, mask.shape)
	return edges

def find_edges_labeled(labels, flat=False):
	"""find_edges_labeled(labels, flat=False)

	Given a 2d numpy array labels[ny,nx], returns an list of indices (y[:],x[:]) for each
	pixel at the edge of a region with constant, nonzero value in labels. If flat==True then it will instead
	return a list of indicees [:] into the flattened labels."""
	# Ensure that we have the right data type and contiguity. Try extra hard to avoid copies
	# when the input is known to have a compatible data type.
	labels = np.asanyarray(labels)
	if labels.dtype in [np.uint32, np.int32]:
		labels = labels.astype(labels.dtype, order="C", copy=False).view(np.int32)
	else:
		labels = labels.astype(np.int32, order="C", copy=False)
	assert labels.ndim == 2, "labels must be 2D"
	cdef inum n = 0
	cdef inum * edges_raw = NULL;
	cdef int[::1] labels_ = labels.reshape(-1)
	cdef inum[::1] view
	n = c.find_edges_labeled(labels.shape[0], labels.shape[1], &labels_[0], &edges_raw)
	# Copy into numpy array
	if n > 0:
		view = <inum[:n]> edges_raw;
		edges = np.array(view)
	else:
		edges = np.zeros(0, np.int64)
	free(edges_raw)
	if not flat:
		edges = np.unravel_index(edges, labels.shape)
	return edges

#### healpix stuff below here ####

# Why is this so verbose?
cdef class healpix_info:
	cdef c.healpix_info * info
	def __init__(self, int nside):
		self.info = c.build_healpix_info(nside)
	def __dealloc__(self):
		c.free_healpix_info(self.info)
	@property
	def nside(self):   return self.info.nside
	@property
	def npix(self):   return self.info.npix
	@property
	def ny(self):      return self.info.ny
	@property
	def nx(self):      return np.asarray(<int[:self.info.ny]> self.info.nx)
	@property
	def off(self):     return np.asarray(<inum[:self.info.ny]> self.info.off)
	@property
	def shift(self):   return np.asarray(<int[:self.info.ny]> self.info.shift)
	@property
	def ra0(self):     return np.asarray(<double[:self.info.ny]> self.info.ra0)
	@property
	def dec(self):     return np.asarray(<double[:self.info.ny]> self.info.dec)
	@property
	def cos_dec(self): return np.asarray(<double[:self.info.ny]> self.info.cos_dec)
	@property
	def sin_dec(self): return np.asarray(<double[:self.info.ny]> self.info.sin_dec)

def find_edges_healpix(healpix_info info, mask, flat=True):
	"""find_edges_healpix(healpix_info info, mask, flat=True)

	Given a healpix_info info and a boolean mask[:], returns
	the pixel indices of all the pixels at the edge of the zero
	region of the mask. If flat=False, then the indices will be
	returned as [{row,col},:]. The default is to return them as
	plain 1d indices."""
	# Flat is the default for healpix because healpix is usually 1d
	cdef int * edges_raw
	cdef inum nedge = 0
	mask = np.asanyarray(mask)
	if mask.dtype in [np.uint8, np.int8, bool]:
		mask = mask.astype(mask.dtype, order="C", copy=False).view(np.uint8)
	else:
		mask = mask.astype(np.uint8, order="C", copy=False)
	assert mask.ndim == 1, "mask must be 1D"
	cdef uint8_t[::1] mask_ = mask
	nedge = c.find_edges_healpix(info.info, &mask_[0], &edges_raw)
	# Copy info numpy array
	cdef int[::1] view = <int[:2*nedge]> edges_raw
	edges = np.array(view).reshape(2,-1)
	free(edges_raw)
	if flat:
		edges = info.off[edges[0]]+edges[1]
	return edges

def find_edges_labeled_healpix(healpix_info info, labels, flat=True):
	"""find_edges_labeled_healpix(healpix_info info, labels, flat=True)

	Given a healpix_info info and integer labels[:], returns
	the pixel indices of all the pixels at the the edge of a
	region with the same, nonzero value in labels. If flat=False,
	then the indices will be returned as [{row,col},:]. The default
	is to return them as plain 1d indices."""
	cdef int * edges_raw
	cdef inum nedge = 0
	labels = np.asanyarray(labels)
	if labels.dtype in [np.uint32, np.int32]:
		labels = labels.astype(labels.dtype, order="C", copy=False).view(np.int32)
	else:
		labels = labels.astype(np.int32, order="C", copy=False)
	assert labels.ndim == 1, "labels must be 2D"
	cdef int[::1] labels_ = labels.reshape(-1)
	nedge = c.find_edges_labeled_healpix(info.info, &labels_[0], &edges_raw)
	# Copy info numpy array
	cdef int[::1] view = <int[:2*nedge]> edges_raw
	edges = np.array(view).reshape(2,-1)
	free(edges_raw)
	if flat:
		edges = info.off[edges[0]]+edges[1]
	return edges

def distance_from_points_healpix(healpix_info info, point_pos, point_pix, rmax=None, omap=None, odomains=None, domains=False, method="bubble"):
	"""distance_from_points_healpix(healpix_info info, point_pos, point_pix, rmax=None, omap=None, odomains=None, domains=False)

	Computes the distance of each healpix pixel in the sky described by healpix_info info to
	the closest of the points with position point_pos[{dec,ra},npoint] and corresponding
	pixel position point_pix[{y,x},npoint]. If omap is specified, the result is written there.
	In any case the result is also returned. If domains=True then a domain map is also returned.
	In this map the value of each pixel is the index of the closest point in the point_pos array,
	or -1 if no valid point was found. The main way this could happen is if rmax is specified,
	which makes the search stop after a distance of rmax radians. Distances larger than this
	will be capped to this value."""
	# Check that our inputs make sense
	point_pos = np.asanyarray(point_pos).astype(float,    order="C", copy=False)
	assert point_pos.ndim == 2 and len(point_pos) == 2, "point_pos must be [{dec,ra},npoint]"
	# point_pix can be standard 1d or our internal 2d format
	point_pix = np.asanyarray(point_pix)
	if point_pix.ndim == 1:
		point_pix = unravel_healpix(info, point_pix)
	point_pix = point_pix.astype(np.int32, order="C", copy=False)
	assert point_pix.ndim == 2 and len(point_pix) == 2 and point_pix.shape[1] == point_pos.shape[1], "point_pos must be [npoint] or [{y,x},npoint]"
	if omap is None: omap = np.empty(info.npix, dtype=np.float64)
	assert omap.ndim == 1 and omap.size == info.npix and omap.dtype==np.float64, "omap must be [npix] float64"
	if odomains is None: odomains = np.empty(info.npix, dtype=np.int32)
	assert odomains.ndim == 1 and odomains.size == info.npix and odomains.dtype==np.int32, "odomains must be [npix] int32"
	# Prepare to call C
	cdef inum npoint = point_pos.shape[1]
	cdef double      rmax_   = (0 if rmax is None else rmax)
	cdef double[::1] point_pos_ = point_pos.reshape(-1)
	cdef int[::1]    point_pix_ = point_pix.reshape(-1)
	cdef double[::1] omap_   = omap.reshape(-1)
	cdef int[::1]    odomains_ = odomains.reshape(-1)
	if method == "bubble":
		c.distance_from_points_bubble_healpix(info.info, npoint, &point_pos_[0], &point_pix_[0], rmax_, &omap_[0], &odomains_[0])
	elif method == "heap":
		c.distance_from_points_heap_healpix(info.info, npoint, &point_pos_[0], &point_pix_[0], rmax_, &omap_[0], &odomains_[0])
	else: raise ValueError("Unrecognized method '%s'" % str(method))
	if domains: return omap, odomains
	else:       return omap

def unravel_healpix(healpix_info info, pix1d):
	pix1d = np.asanyarray(pix1d).astype(np.int64, order="C", copy=False)
	pix2d = np.empty((2,)+pix1d.shape, np.int32)
	cdef npoint = pix1d.size
	cdef inum[::1] pix1d_ = pix1d.reshape(-1)
	cdef int[::1]  pix2d_ = pix2d.reshape(-1)
	c.unravel_healpix(info.info, npoint, &pix1d_[0], &pix2d_[0])
	return pix2d

def ravel_healpix(healpix_info info, pix2d):
	pix2d = np.asanyarray(pix2d).astype(np.int32, order="C", copy=False)
	pix1d = np.empty(pix2d.shape[1:], np.int64)
	cdef npoint = pix1d.size
	cdef int[::1]  pix2d_ = pix2d.reshape(-1)
	cdef inum[::1] pix1d_ = pix1d.reshape(-1)
	c.ravel_healpix(info.info, npoint, &pix2d_[0], &pix1d_[0])
	return pix1d

def get_healpix_neighs(healpix_info info, int y, int x):
	res = np.zeros([2,4],np.int32)
	cdef int[::1] yneigh = res[0]
	cdef int[::1] xneigh = res[1]
	c.get_healpix_neighs(info.info, y, x, &yneigh[0], &xneigh[0])
	return res



"""This module provides functions that make it easier to write curvature-agnostic code
that looks the same whether it's operating using the flat-sky approximation (2d FFTs)
or the curved sky (SHTs and alms)"""
import numpy as np
from . import utils, enmap, curvedsky

# Unified Harmonic Transform
class UHT:
	"""The UHT class provides a Unified Harmonic Transform, which provides both
	FFTs and SHTs under a unified interface. The purpose of this class is to be able
	to write various filtering and convolution operations in a flat-sky-agnostic way.
	
	After initialization, the main purpose of the UHT object is to transform between
	pixel-space ("map") and harmonic-space ("harm") representations. In both flat-sky
	and curved-sky modes, the map-space representation is an enmap, but the harmonic
	space reprsentation differ for the two. In flat-sky the harmonic representation is
	also an enmap, while in curved-sky mode it's an alm.

	The UHT object also provides functions for working with 1d functions or r or l.
	Functions of r can be transformed to harmonic form using rprof2hprof. Due to
	the symmetry, the harmonic space reprsentation of these is a bit different from
	that of the maps. In flat mode, the result is still an enmap, but in curved mode
	it's just a 1d function of l. If you already have a function of l, it can be put
	in the standard form using the lexpand function.

	You can multiply the harmonic representation of profiles with that of the maps
	by using lmul.

	This all sounds complicated, but the point is that this UHT object takes care
	of these details. For example, here is how you would smooth a map given the
	real-space reprsentation of a beam:

		uht  = UHT(shape, wcs)
		beam = uth.rprof2hprof(br, r)
		omap = uht.harm2map(uht.lmul(beam, uht.map2harm(map)))

	Notice how none of this depended on flat/curved sky or the the details of the harmonic
	space representation.

	Explanation of terms and representations:

	* map:   An enmap with shape [...,ny,nx] representing some or all of the sky
	* rprof: Radial profiles br[...,nr] valuated at the points r[nr], typically a beam.
	* lprof: A 1d function of l: bl[...,nl]. Does not have to match the internal lmax.
	* harm:  The harmonic-space representation of a map, as produced by map2harm.
	  - flat:   A complex enmap[...,ny,nx] corresponding to the the multipoles self.l
	  - curved: Alms[...,self.lmax+1]
	* hprof: The harmonic-space representation of an isotropic function, as produced by
	  rprof2hprof or lprof2hprof. Can be multiplied with harm using lmul.
	  - flat:   An enmap[...,ny,nx] corresponding to the multipoles self.l
	  - curved: A 1d array [...,self.lmax+1]
	"""
	def __init__(self, shape, wcs, mode="auto", lmax=None, max_distortion=0.1, niter=0):
		"""Initialize a UHT object.
		Arguments:
			shape, wcs: The geometry of the enmaps the UHT object will be used on.
			mode: The flat/curved sky mode to use. "flat" selects the flat-sky approximation,
				"curved" uses full SHTs, and "auto" (the detault) selects flat or curved based on
				the estimated maximum distortion in the map.
			lmax: The maximum multipole to use in calculations. Ignored in flat-sky mode.
			max_distortion: The maximum relative scale difference across the map that's acceptable
				before curved sky is necessary."""
		self.shape, self.wcs = shape[-2:], wcs
		self.area = enmap.area(self.shape, self.wcs)
		self.fsky = self.area/(4*np.pi)
		if mode == "auto":
			dist = estimate_distortion(shape, wcs)
			if dist <= max_distortion: mode = "flat"
			else:                      mode = "curved"
		self.mode = mode
		self.quad = None
		self.niter= niter
		if mode == "flat":
			self.l    = enmap.modlmap(shape, wcs)
			self.lmax = utils.nint(np.max(self.l))
			# FIXME: In the middle of debugging this. Was
			# nper = l[0,1]*l[1,0]/pi = 1/fsky, but was inconsistent with sum_harm for curved
			self.nper = 1/self.fsky
			self.ntot = self.nper*self.shape[-2]*self.shape[-1]
		elif mode == "curved":
			if lmax is None:
				res  = np.min(np.abs(wcs.wcs.cdelt))*utils.degree
				lmax = res2lmax(res)
			self.lmax = lmax
			self.l    = np.arange(lmax+1)
			self.ainfo= curvedsky.alm_info(lmax=lmax)
			self.nper = 2*self.l+1
			self.ntot = np.sum(self.nper)
		else:
			raise ValueError("Unrecognized mode in UHT: '%s'" % (str(mode)))
	@property
	def npix(self): return self.shape[-2]*self.shape[-1]
	def map2harm(self, map, spin=0):
		if self.mode == "flat":
			return enmap.map2harm(map, spin=spin, normalize="phys")
		else:
			return curvedsky.map2alm(map, ainfo=self.ainfo, spin=spin, niter=self.niter)
	def harm2map(self, harm, spin=0):
		if self.mode == "flat":
			return enmap.harm2map(harm, spin=spin, normalize="phys").real
		else:
			rtype= np.zeros(1, harm.dtype).real.dtype
			omap = enmap.zeros(harm.shape[:-1]+self.shape, self.wcs, rtype)
			return curvedsky.alm2map(harm, omap, ainfo=self.ainfo, spin=spin)
	def harm2map_adjoint(self, map, spin=0):
		if self.mode == "flat":
			return enmap.harm2map_adjoint(map, spin=spin, normalize="phys")
		else:
			return curvedsky.alm2map_adjoint(map, ainfo=self.ainfo)
	def map2harm_adjoint(self, harm, spin=0):
		if self.mode == "flat":
			return enmap.map2harm_adjoint(harm, spin=spin, normalize="phys")
		else:
			rtype= np.zeros(1, harm.dtype).real.dtype
			omap = enmap.zeros(harm.shape[:-1]+self.shape, self.wcs, rtype)
			return curvedsky.map2alm_adjoint(harm, omap, ainfo=self.ainfo, spin=spin, niter=self.niter)
	def quad_weights(self):
		"""Returns the quadrature weights array W. This will broadcast correctly
		with maps, but may not have the same dimensions due to symmetries.
		map2harm = harm2map_adjoint * quad_weight"""
		if self.quad is None:
			if self.mode == "flat":
				self.quad = enmap.pixsizemap(self.shape, self.wcs, broadcastable=True)
			else:
				self.quad = curvedsky.quad_weights(self.shape, self.wcs)[:,None]
		return self.quad
	def rprof2hprof(self, br, r):
		"""Like map2harm, but for a 1d function of r."""
		if self.mode == "flat":
			return profile2harm_flat_2d(br, r, self.shape, self.wcs)
		else:
			return curvedsky.profile2harm(br, r, lmax=self.lmax)
	def hprof2rprof(self, harm, r):
		"""Inverse of hprof2rprof"""
		if self.mode == "flat":
			return harm2profile_flat_2d(harm+0j, r)
		else:
			return curvedsky.harm2profile(harm, r)
	def lprof2hprof(self, lprof):
		if self.mode == "flat":
			return enmap.enmap(utils.interpol(lprof, self.l[None], order=1, border="constant"), self.wcs, copy=False)
		else:
			if lprof.shape[-1] >= self.lmax+1:
				return lprof[...,:self.lmax+1]
			else:
				return np.concatenate([lprof, np.zeros(lprof.shape[:-1]+(self.lmax+1-lprof.shape[-1],), lprof.dtype)], -1)
	def hprof2harm(self, hprof):
		if self.mode == "flat":
			return hprof.copy()
		else:
			mapping = self.ainfo.get_map()
			return hprof[...,mapping[:,0]]
	def hmul(self, hprof, harm, inplace=False):
		"""Perform the multiplication hprof*harm -> harm. See the UHT class docstring for
		the meaning of these terms. In flat mode, hprof must be [ny,nx], [ncomp,ny,nx] or
		[ncomp,ncomp,ny,nx].  In curved mode, hprof must be [nl], [ncomp,nl] or [ncomp,ncomp,nl]."""
		harm = np.asanyarray(harm)
		if self.mode == "flat":
			if inplace: harm[:] = enmap.map_mul(hprof, harm)
			else:       harm    = enmap.map_mul(hprof, harm)
			return harm
		else:
			out = harm if inplace else None
			harm = harm.astype(np.result_type(harm,0j), copy=False)
			return self.ainfo.lmul(harm, hprof, out=out)
	def hrand(self, hprof):
		"""Draw a random realization of the harmonic profile hprof"""
		if self.mode == "flat":
			noise = enmap.rand_gauss_harm(self.shape, self.wcs)
			return enmap.map_mul(enmap.multi_pow(hprof/noise.pixsize(),0.5), noise)
		else:
			return curvedsky.rand_alm(hprof, lmax=self.lmax)
	def harm2powspec(self, harm, harm2=None, patch=False):
		"""Compute the pseudo power spectrum corresponding to harm, or a cross-spectrum if harm2 is given.
		Return an hprof. If patch == True, then the harm is assumed to have been measured from
		map2harm, and an fsky correction will be applied in the curved-sky case."""
		if self.mode == "flat":
			return enmap.calc_ps2d(harm, harm2)
		else:
			powspec = curvedsky.alm2cl(harm, harm2)
			if patch: powspec /= self.fsky
			return powspec
	def sum_hprof(self, hprof):
		"""Sum the total signal in a harmonic profile, typically a power spectrum"""
		hprof = np.asanyarray(hprof)
		if self.mode == "flat":
			return np.sum(hprof*self.nper,(-2,-1))
		else:
			return np.sum(hprof*self.nper,-1)
	def mean_hprof(self, hprof): return self.sum_hprof(hprof)/self.ntot
	def hprof_rpow(self, hprof, power):
		"""Raises hprof "hprof" to the power "power" in real-space.
		Effectively map2harm(harm2map(hprof)**power), but works on
		harmonic profiles instead of full fourier maps/alms.
		Maybe this function is too specific to have in this class, but
		I needed it and it wasn't trivial to get right.
		"""
		if self.mode == "flat":
			norm = enmap.area(self.shape, self.wcs)**0.5
			map  = self.harm2map(hprof/norm+0j)
			return self.map2harm(map**power)*norm
		else:
			# Estimate the resolution from the beam
			sigma = 1/max(1,np.where(hprof > np.max(hprof)*np.exp(-0.5))[0][-1])
			r     = np.arange(0, 20*sigma, sigma/10)
			rprof = self.hprof2rprof(hprof, r)
			return self.rprof2hprof(rprof**power, r)

####################
# Helper functions #
####################

def profile2harm_flat(br, r, oversample=2, pad_factor=2):
	"""Flat-sky approximation to curvedsky.profile2harm. Accurate to about 0.5% for a 1.4 arcmin
	fwhm beam. Only supports a 1d br"""
	# Build a 2d pixelization we will evaluate the ffts on
	res  = beam2res(br, r)
	rmax = beam2rmax(br, r)*pad_factor
	n    = 2*utils.nint(rmax/res*oversample)+1
	shape, wcs = enmap.geometry(pos=[0,0], res=res/oversample, shape=(n,n), proj="car")
	# Compute the 2d beam in harmonic space
	lbeam_2d = profile2harm_flat_2d(br, r, shape, wcs)
	# Reduce to equispaced standard beam
	bl_tmp, l_tmp = lbeam_2d.lbin()
	lmax = res2lmax(res)
	l  = np.arange(lmax+1)
	bl = np.interp(l, l_tmp, bl_tmp)
	return bl

def profile2harm_flat_2d(br, r, shape, wcs):
	"""Given a 1d beam br(r), compute the 2d beam transform bl(ly,lx) for
	the l-space of the map with the given shape, wcs, assuming a flat sky.
	Despite the name, it is not specific to beams - any real-space function of r
	will do. br can have arbitrary pre-dimensions [...,nr]"""
	br     = np.asarray(br)
	cpix   = np.array(shape[-2:])//2-1
	cpos   = enmap.pix2sky(shape, wcs, cpix)
	rmap   = enmap.shift(enmap.modrmap(shape, wcs, cpos), -cpix)
	bmap   = enmap.ndmap(utils.interp(rmap, r, br, right=0), wcs)
	# Normalize the beam so that l=0 corresponds to the sky mean of the beam,
	# like it is for get_lbeam_exact
	harm  = enmap.fft(bmap, normalize=False).real
	harm *= harm.pixsize()
	return harm

def harm2profile_flat_2d(harm, r=None):
	"""Inverse of profile2harm_flat_2d. harm should be an enmap.
	r is [:] in radians."""
	bmap = enmap.ifft(harm, normalize=False).real
	bmap/= harm.pixsize() * harm.npix
	cpix = np.array(harm.shape[-2:])//2-1
	cpos = bmap.pix2sky(cpix)
	bmap = enmap.shift(bmap, cpix, keepwcs=True)
	wbr, wr = bmap.rbin(center=cpos)
	if r is None: return wbr, r
	else:         return utils.interp(r, wr, wbr, right=0)

def beam2res(br, r):
	fwhm = 2*r[np.where(br>=br[0]*0.5)[0][-1]]
	res  = fwhm/3
	return res

def beam2rmax(br, r, tol=1e-5, return_index=False):
	imax = np.where(br>=br[0]*tol)[0][-1]
	if return_index: return r[imax], imax
	else:            return r[imax]

def res2lmax(res):
	"""Get the lmax needed to represent the spatial scale res in radians"""
	return utils.nint(np.pi/res)

def estimate_distortion(shape, wcs):
	"""Get the maximum distortion in the map, assuming a cylindrical projection"""
	dec1, dec2 = enmap.corners(shape, wcs)[:,0]
	rmin = min(np.cos(dec1),np.cos(dec2))
	rmax = 1 if not dec1*dec2 > 0 else max(np.cos(dec1),np.cos(dec2))
	return rmax/rmin-1



import numpy as np, time, contextlib
from . import bunch

class Device:
	def __init__(self):
		self.pools = None # Memory pools
		self.np    = None # numpy or equivalent
		self.lib   = bunch.Bunch() # place to store library functions
	def get(self, arr): raise NotImplementedError # copy device array to cpu
	def ptr(self, arr): return getptr(arr)
	def synchronize(self): raise NotImplementedError
	def garbage_collect(self): raise NotImplementedError
	def memuse(self, type="total"): raise NotImplementedError
	def copy(self, afrom, ato): raise NotImplementedError
	def time(self):
		self.synchronize()
		return time.time()

class DeviceCpu(Device):
	def __init__(self, align=None, alloc_factory=None, verbose=False):
		super().__init__()
		if align is None: align = 16
		if alloc_factory is None:
			def alloc_factory(name):
				return ArrayPoolCpu(AllocAligned(AllocCpu(), align=align), name=name, verbose=verbose)
		self.pools = ArrayMultipool(alloc_factory)
		self.np    = np
	def get(self, arr): return arr.copy()
	def synchronize(self): pass
	def garbage_collect(self):
		import gc
		gc.collect()
	def memuse(self, type="total"):
		if type == "total":
			from . import memory
			return memory.current()
		elif type == "pools":
			return self.pools.capacity()
		elif type == "np":
			return 0
		else: raise ValueError("Unknown memuse type: '%s'" % str(type))
	def copy(self, afrom, ato):
		"""Copy (cpu/dev) → (cpu/dev)"""
		ato[:] = afrom

class DeviceGpu(Device):
	def __init__(self, align=None, alloc_factory=None, verbose=False):
		super().__init__()
		if align is None: align = 512
		import cupy
		if alloc_factory is None:
			def alloc_factory(name):
				return ArrayPoolGpu(AllocAligned(AllocGpu(), align=align), name=name, verbose=verbose)
		self.pools = ArrayMultipool(alloc_factory)
		self.np    = cupy
		self.heap  = cupy.get_default_memory_pool()
		self.nvhandle = None
	def get(self, arr): return arr.get()
	def synchronize(self):
		import cupy
		cupy.cuda.runtime.deviceSynchronize()
	def garbage_collect(self):
		self.heap.free_all_blocks()
	def memuse(self, type="total"):
		if type == "total":
			import nvidia_smi
			if self.nvhandle is None:
				nvidia_smi.nvmlInit()
				self.nvhandle = nvidia_smi.nvmlDeviceGetHandleByIndex(0)
			info   = nvidia_smi.nvmlDeviceGetMemoryInfo(self.nvhandle)
			return info.used
		elif type == "pools":
			return self.pools.capacity()
		elif type == "np":
			return self.heap.used_bytes()
		else: raise ValueError("Unknown memuse type: '%s'" % str(type))
	def copy(self, afrom, ato):
		"""Copy (cpu/dev) → (cpu/dev)"""
		cuda_memcpy(afrom,ato)

# I'm having trouble using these allocators in place of
# the built-in ones. I think wrapping everything in arrays
# from the beginning is messing things up. The basic allocators
# should simply deal with MemoryPointers, and arrays should only
# enter at the ArrayPool level. To make this work, I need a version
# of MemoryPointer for the cpu too. The gpu version has:
#  .mem: PooledMemory, which has .size
#  .ptr: int, pointer to our part
#  etc.
# My code uses this to construct cupy arrays with
# cupy.ndarray(shape, dtype, memptr=...)
#
# Numpy is missing:
# * allocator support
# * suport for building array from pointer
# Maybe it's not worth it to try to make these look the same
# at this low level. If I get rid of AllocAligned, then
# all the code is CPU/GPU-specific

class AllocCpu:
	def alloc(self, n): return np.empty(n, dtype=np.uint8)

class AllocGpu:
	def __init__(self):
		import cupy
		self.allocator = cupy.cuda.get_allocator()
	def alloc(self, n):
		import cupy
		n      = int(n)
		memptr = self.allocator(n)
		return cupy.ndarray(n, np.uint8, memptr=memptr)

class AllocAligned:
	"""Wraps an allocator to make it aligned. Should work for both cpu and gpu.
	A bit inefficient if the underlying allocator is already aligned, which it
	probably already is."""
	def __init__(self, allocator, align=16):
		self.allocator = allocator
		self.align     = align
	def alloc(self, n):
		n   = int(n)
		buf = self.allocator.alloc(n+self.align-1)
		off = (-getptr(buf))%self.align
		return buf[off:off+n]

class Mempool:
	def __init__(self, aligned_alloc, name="[unnamed]", verbose=False):
		self.allocator = aligned_alloc
		self.name      = name
		self.verbose   = verbose
		self.free()
	def alloc(self, n):
		n       = int(n)
		effsize = round_up(n, self.allocator.align)
		# We have two modes:
		# 1. hand out memory from a single existing allocation.
		#    This is only active when len(arenas) == 1
		# 2. append new arenas
		if len(self.arenas) != 1 or self.arenas[0].size < self.used + n:
			if self.verbose:
				print("Growing pool %s to size %d by %d. %s" % (self.name, self.used + n, n, str([a.size for a in self.arenas])))
			# We're in mode 1
			self.arenas.append(self.allocator.alloc(n))
			buf = self.arenas[-1][0:n]
			self.used += effsize
		else:
			# We're in mode 2. Hand out some more memory
			buf        = self.arenas[-1][self.used:self.used+n]
			self.used += effsize
		return buf
	@property
	def capacity(self): return self.arenas[0].size if len(self.arenas) == 1 else self.used
	def free(self):
		self.arenas = []
		self.used   = 0
	def reset(self):
		"""Invalidate the memory we point to, potentially cleaning it up to a single
		fixed area we can allocate from. New allocations will reuse this memory
		as long as they don't exceed its capacity. If the capacity is exceeded, then
		it will start requesting new memory again."""
		if len(self.arenas) != 1:
			# Go to from mode 1 to mode 2
			self.arenas = [self.allocator.alloc(self.capacity)]
		self.used = 0
		return self
	def reserve(self, n):
		"""Reserve space for at least n bytes without reallocation"""
		self.reset()  # Reset position
		self.alloc(n) # Perform an allocation
		self.reset()  # Mark it as unused
		return self
	def __repr__(self):
		arenas = "["+",".join(["%.3fG" % (arena.size/1024**3) for arena in self.arenas])+"]"
		return "%s(name='%s', capacity=%.3fG, used=%.3fG, align=%d, arenas=%s)" % (self.__class__.__name__, self.name, self.capacity/1024**3, self.used/1024**3, self.allocator.align, arenas)

class ArrayPoolCpu(Mempool):
	def array(self, arr, reset=True, verbose=False):
		self.verbose = verbose
		arr  = np.asarray(arr)
		oarr = self.empty(arr.shape, dtype=arr.dtype, reset=True)
		oarr[:] = arr
		return oarr
	def empty(self, shape, dtype=np.float32, reset=True):
		if reset: self.reset()
		return self.alloc(np.prod(shape)*np.dtype(dtype).itemsize).view(dtype).reshape(shape)
	def full(self, shape, val, dtype=np.float32, reset=True):
		arr = self.empty(shape, dtype=dtype, reset=reset)
		arr[:] = val
		return arr
	def zeros(self, shape, dtype=np.float32, reset=True):
		return self.full(shape, 0, dtype=dtype, reset=reset)
	def ones(self, shape, dtype=np.float32, reset=True):
		return self.full(shape, 1, dtype=dtype, reset=reset)
	def copy(self, other, reset=True):
		arr = self.empty(other.shape, other.dtype, reset=reset)
		arr[:] = other
		return arr
	# No allocator support in numpy, so undefined what this
	# should return. Just return a numpy array of bytes for now
	def alloc_raw(self, n): return self.alloc(n)
	@contextlib.contextmanager
	def as_allocator(self, reset=True):
		if reset: self.reset()
		# No-op
		try: yield
		finally: pass

class ArrayPoolGpu(Mempool):
	def array(self, arr, reset=True, verbose=False):
		# Make sure the array is contiguous, which our memcpy needs
		import cupy
		self.verbose = verbose
		ap   = cupy if isinstance(arr, cupy.ndarray) else np
		arr  = ap.ascontiguousarray(arr)
		oarr = self.empty(arr.shape, dtype=arr.dtype, reset=True)
		cuda_memcpy(arr, oarr)
		return oarr
	def empty(self, shape, dtype=np.float32, reset=True):
		if reset: self.reset()
		return self.alloc(np.prod(shape)*np.dtype(dtype).itemsize).view(dtype).reshape(shape)
	def full(self, shape, val, dtype=np.float32, reset=True):
		arr = self.empty(shape, dtype=dtype, reset=reset)
		arr[:] = val
		return arr
	def zeros(self, shape, dtype=np.float32, reset=True):
		return self.full(shape, 0, dtype=dtype, reset=reset)
	def ones(self, shape, dtype=np.float32, reset=True):
		return self.full(shape, 1, dtype=dtype, reset=reset)
	def copy(self, other, reset=True):
		arr = self.empty(other.shape, other.dtype, reset=reset)
		arr[:] = other
		return arr
	def alloc_raw(self, n): return self.alloc(n).data
	@contextlib.contextmanager
	def as_allocator(self, reset=True):
		import cupy
		if reset: self.reset()
		old_allocator = cupy.cuda.get_allocator()
		try:
			# This causes a crash. I think it happens because
			# reset() can end up freeing the memory we handed
			# out here. In the old design, things were done a bit
			# differently. The memory needed was pre-allocated
			# and parseled out. However, if a later to was bigger than
			# an earlier, then memory would still be reallocated, so I
			# don't see why that worked.
			# TODO: Make a small test case that demonstrates the problem
			# without all this abstraction.
			cupy.cuda.set_allocator(self.alloc_raw)
			yield
		finally:
			cupy.cuda.set_allocator(old_allocator)

class ArrayMultipool:
	def __init__(self, factory):
		self.factory = factory
		self.pools   = {}
	def want(self, *names):
		pools = []
		for name in names:
			if name not in self.pools:
				self.pools[name] = self.factory(name=name)
			pools.append(self.pools[name])
		return pools
	def used(self): return sum([pool.used for name, pool in self.pools.items()])
	def capacity(self): return sum([pool.capacity for name, pool in self.pools.items()])
	def free(self):
		for name in self.pools:
			self.pools[name].free()
	def reset(self):
		for name in self.pools:
			self.pools[name].reset()
	def swap(self, name1, name2):
		self.pools[name1], self.pools[name2] = self.pools[name2], self.pools[name1]
	def __getitem__(self, name):
		"""Returns the memory pool with the given name, creating it if it doesn't exist"""
		if name not in self.pools:
			self.pools[name] = self.factory(name=name)
		return self.pools[name]
	def __getattr__(self, name):
		"""Returns the memory pool with the given name, if it exists"""
		return self.pools[name]
	def __delattr__(self, name):
		del self.pools[name]
	def __repr__(self):
		msg = "ArrayMultipool("
		names = sorted(list(self.pools.keys()))
		for name in names:
			msg += "\n  %s" % repr(self.pools[name])
		if len(names) > 0:
			msg += "\n"
		msg += ")"
		return msg

def round_up(a,b): return (a+b-1)//b*b

def cuda_memcpy(afrom,ato):
	import cupy
	assert afrom.flags["C_CONTIGUOUS"] and ato.flags["C_CONTIGUOUS"]
	cupy.cuda.runtime.memcpy(getptr(ato), getptr(afrom), ato.nbytes, cupy.cuda.runtime.memcpyDefault)

def getptr(arr):
	"""Returns a pointer to arr's data whether it's a numpy or cupy array"""
	try: return arr.data.ptr
	except AttributeError: return arr.ctypes.data

def anypy(arr):
	"""Returns the numpy or cupy modules depending on whether arr is
	a numpy or cupy array, or can be converted to these. Raises a
	ValueError if neither is the case. Still works if cupy can't be
	imported, as long as arr isn't a cupy array"""
	try: import cupy
	except ModuleNotFoundError: return np
	try:
		np.asanyarray(arr)
		return np
	except TypeError: pass
	try:
		cupy.asanyarray(arr)
		return cupy
	except TypeError: pass
	raise ValueError("Neither numpy or cupy array, and can't be converted to them either")



import numpy as np, copy

def _unbuf(bufspec):
    return bufspec[0] if isinstance(bufspec, tuple) else np.asarray(bufspec)

class FakeCommunicator:
    def __init__(self):
        self.size = 1
        self.rank = 0
    def Allreduce(self, sendbuf, recvbuf, op=lambda a,b:a+b):
        _unbuf(recvbuf)[()] = _unbuf(sendbuf)
    def Allgather(self, sendbuf, recvbuf):
        _unbuf(recvbuf)[0] = _unbuf(sendbuf)
    def Allgatherv(self, sendbuf, redvbuf):
        _unbuf(recvbuf)[()] = _unbuf(sendbuf)
    def Alltoallv(self, sendbuf, recvbuf):
        _unbuf(recvbuf)[()] = _unbuf(sendbuf)
    def Barrier(self): pass
    def allreduce(self, sendobj, op=lambda a,b:a+b):
        return copy.deepcopy(sendobj)
    def allgather(self, sendobj, op=lambda a,b:a+b):
        return [copy.deepcopy(sendobj)]
    def allgatherv(self, sendobj, op=lambda a,b:a+b):
        return [copy.deepcopy(sendobj)]

FAKE_WORLD = FakeCommunicator()



from libc.stdint cimport int64_t
cdef extern from "cmisc_core.h":
	void alm2cl_sp(int lmax, int mmax, int64_t * mstart, float  * alm1, float  * alm2, float * cl)
	void alm2cl_dp(int lmax, int mmax, int64_t * mstart, double * alm1, double * alm2, double * cl)
	void transpose_alm_dp(int lmax, int mmax, int64_t * mstart, double * ialm, double * oalm)
	void transpose_alm_sp(int lmax, int mmax, int64_t * mstart, float * ialm, float * oalm)
	void lmul_dp(int lmax, int mmax, int64_t * mstart, double * alm, int lfmax, const double * lfun)
	void lmul_sp(int lmax, int mmax, int64_t * mstart, float * alm, int lfmax, const float * lfun)
	void transfer_alm_dp(int lmax1, int mmax1, int64_t * mstart1, double * alm1, int lmax2, int mmax2, int64_t * mstart2, double * alm2)
	void transfer_alm_sp(int lmax1, int mmax1, int64_t * mstart1, float * alm1, int lmax2, int mmax2, int64_t * mstart2, float * alm2)
	void lmatmul_dp(int N, int M, int lmax, int mmax, int64_t * mstart, double ** alm, int lfmax, double ** lmat, double ** oalm)
	void lmatmul_sp(int N, int M, int lmax, int mmax, int64_t * mstart, float ** alm, int lfmax, float ** lmat, float ** oalm)



import time
from contextlib import contextmanager
from . import bunch

"""bench: Simple timing of python code blocks.

Example usage:

	1. Manual printing

	from pixell import bench
	for i in range(nfile):
		with bench.mark("all"):
			with bench.mark("read"):
				a = np.loadtxt(afiles[i])
				b = np.loadtxt(bfiles[i])
			with bench.mark("sum"):
				a += b
			with bench.mark("write"):
				np.savetxt(ofiles[i], a)
		print("Processed case %d in %7.4f s. read %7.4f sum %7.4f write %7.4f" % (i, bench.t.all, bench.t.read, bench.t.sum, bench.t.write))
	print("Total %7.4f s. read %7.4f sum %7.4f write %7.4f" % (i, bench.t_tot.all, bench.t_tot.read, bench.t_tot.sum, bench.t_tot.write))

	2. Quick-and-dirty printing

	from pixell import bench
	for i in range(nfile):
		with bench.show("read"):
			a = np.loadtxt(afiles[i])
			b = np.loadtxt(bfiles[i])
		with bench.show("sum"):
			a += b
		with bench.show("write"):
			np.savetxt(ofiles[i], a)

	bench.show is equivalent to bench.mark, just with an extra print.
	This means that bench.show updates .ttot and .n just like bench.mark
	does.

The examples above collect statistics globally. You can create local
benchmark objects with bench.Bench(). Example:

	from pixell import bench
	mybench = bench.Bench()
	with mybench.mark("example"):
		do_something()

The overhead of bench.mark is around 3 µs.
"""

_print   = print

# Just wall times for now, but could be extended to measure
# cpu time or leaked memory
class Bench:
	def __init__(self, verbose=False):
		self.t      = bunch.Bunch()
		self.t_tot  = bunch.Bunch()
		self.n      = bunch.Bunch()
		self.verbose = verbose
	@contextmanager
	def mark(self, name, tfun=None):
		if tfun is None: tfun = time.time
		t1 = tfun()
		try:
			yield
		finally:
			t2 = tfun()
			self.add(name, t2-t1)
			if self.verbose:
				self.print(name)
	@contextmanager
	def show(self, name, tfun=None):
		try:
			with self.mark(name, tfun=None):
				yield
		finally:
			self.print(name)
	def add(self, name, t):
		if name not in self.n:
			self.t_tot[name] = 0
			self.n    [name] = 0
		self.n    [name] += 1
		self.t    [name]  = t
		self.t_tot[name] += t
	def print(self, name):
		_print("%7.4f s (last) %7.4f s (mean) %4d (n) %s" % (self.t[name], self.t_tot[name]/self.n[name], self.n[name], name))
	def set_verbose(self, verbose):
		self.verbose = verbose

# Global interface
_default = Bench()
mark  = _default.mark
show  = _default.show
add   = _default.add
print = _default.print
t_tot = _default.t_tot
t     = _default.t
n     = _default.n
set_verbose = _default.set_verbose



import subprocess
import multiprocessing
import sys
from pathlib import Path


def main():
    max_threads = multiprocessing.cpu_count()
    assert max_threads >= 1

    def run_benchmark(nthreads):
        subprocess.call(
            [sys.executable, Path(__file__).parent / "benchmark_pixell_runner.py"],
            env={"OMP_NUM_THREADS": str(nthreads)},
        )

    print("Single threaded alm test:")
    run_benchmark(1)

    print(f"Multi-threaded alm test with {max_threads} threads:")
    run_benchmark(max_threads)
